# -*- coding: utf-8 -*-
"""Dependencies: flask, tornado."""
import logging
from wbia.control import accessor_decors, controller_inject
from wbia import constants as const
import utool as ut
import simplejson as json
from os.path import join, dirname, abspath, exists
from flask import url_for, request, current_app
from wbia.constants import KEY_DEFAULTS, SPECIES_KEY
from wbia.web import appfuncs as appf
import numpy as np

(print, rrr, profile) = ut.inject2(__name__)
logger = logging.getLogger('wbia')

CLASS_INJECT_KEY, register_ibs_method = controller_inject.make_ibs_register_decorator(
    __name__
)
register_api = controller_inject.get_wbia_flask_api(__name__)
register_route = controller_inject.get_wbia_flask_route(__name__)


@register_ibs_method
@accessor_decors.getter_1toM
@register_api('/api/wic/cnn/', methods=['PUT', 'GET', 'POST'])
def wic_cnn(ibs, gid_list, testing=False, algo='cnn', model_tag='candidacy', **kwargs):
    depc = ibs.depc_image
    config = {}

    if model_tag is not None:
        config['classifier_two_algo'] = algo
        config['classifier_two_weight_filepath'] = model_tag

    if testing:
        depc.delete_property('classifier_two', gid_list, config=config)

    result_list = depc.get_property('classifier_two', gid_list, None, config=config)

    output_list = []
    for result in result_list:
        scores, classes = result
        output_list.append(scores)

    return output_list


@register_ibs_method
@accessor_decors.getter_1to1
def wic_cnn_json(ibs, gid_list, config={}, **kwargs):
    return wic_cnn(ibs, gid_list, **config)


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/detect/randomforest/', methods=['PUT', 'GET'])
def detect_random_forest(ibs, gid_list, species, commit=True, **kwargs):
    """Run animal detection in each image. Adds annotations to the database as they are found.

    Args:
        gid_list (list): list of image ids to run detection on
        species (str): string text of the species to identify

    Returns:
        aids_list (list): list of lists of annotation ids detected in each
            image

    CommandLine:
        python -m wbia.web.apis_detect --test-detect_random_forest --show

    RESTful:
        Method: PUT, GET
        URL:    /api/detect/randomforest/

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.web.apis_detect import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb('testdb1')
        >>> gid_list = ibs.get_valid_gids()[0:2]
        >>> species = wbia.const.TEST_SPECIES.ZEB_PLAIN
        >>> aids_list = ibs.detect_random_forest(gid_list, species)
        >>> # Visualize results
        >>> if ut.show_was_requested():
        >>>     import wbia.plottool as pt
        >>>     from wbia.viz import viz_image
        >>>     for fnum, gid in enumerate(gid_list):
        >>>         viz_image.show_image(ibs, gid, fnum=fnum)
        >>>     pt.show_if_requested()
        >>> # Remove newly detected annotations
        >>> ibs.delete_annots(ut.flatten(aids_list))
    """
    # TODO: Return confidence here as well
    depc = ibs.depc_image
    config = {
        'algo': 'rf',
        'species': species,
        'sensitivity': 0.2,
        'nms': True,
        'nms_thresh': 0.4,
    }
    results_list = depc.get_property('localizations', gid_list, None, config=config)
    if commit:
        aids_list = ibs.commit_localization_results(
            gid_list, results_list, note='pyrfdetect'
        )
        return aids_list

    # results_list = depc.get_property('detections', gid_list, None, config=config)
    # if commit:
    #     aids_list = ibs.commit_detection_results(gid_list, results_list, note='pyrfdetect')
    #     return aids_list


@register_route('/test/review/detect/cnn/yolo/', methods=['GET'])
def review_detection_test(
    image_uuid=None,
    result_list=None,
    callback_url=None,
    callback_method='POST',
    **kwargs,
):
    ibs = current_app.ibs
    if image_uuid is None or result_list is None:
        results_dict = ibs.detection_yolo_test()
        image_uuid = results_dict['image_uuid_list'][0]
        result_list = results_dict['results_list'][0]
    if callback_url is None:
        callback_url = request.args.get('callback_url', url_for('process_detection_html'))
    if callback_method is None:
        callback_method = request.args.get('callback_method', 'POST')
    template_html = review_detection_html(
        ibs, image_uuid, result_list, callback_url, callback_method, include_jquery=True
    )
    template_html = """
        <script src="http://code.jquery.com/jquery-2.2.1.min.js" ia-dependency="javascript"></script>
        %s
    """ % (
        template_html,
    )
    return template_html


@register_ibs_method
@register_api('/test/detect/cnn/yolo/', methods=['GET'])
def detection_yolo_test(ibs, config={}):
    from random import shuffle  # NOQA

    gid_list = ibs.get_valid_gids()
    shuffle(gid_list)
    gid_list = gid_list[:3]
    results_dict = ibs.detect_cnn_yolo_json(gid_list, config=config)
    return results_dict


@register_ibs_method
@register_api('/test/detect/cnn/lightnet/', methods=['GET'])
def detection_lightnet_test(ibs, config={}):
    from random import shuffle  # NOQA

    gid_list = ibs.get_valid_gids()
    shuffle(gid_list)
    gid_list = gid_list[:3]
    results_dict = ibs.detect_cnn_lightnet_json(gid_list, config=config)
    return results_dict


@register_api('/api/review/detect/cnn/yolo/', methods=['GET'])
def review_detection_html(
    ibs,
    image_uuid,
    result_list,
    callback_url,
    callback_method='POST',
    include_jquery=False,
    config=None,
):
    """
    Return the detection review interface for a particular image UUID and a list of results for that image.

    Args:
        image_uuid (UUID): the UUID of the image you want to review detections for
        result_list (list of dict): list of detection results returned by the detector
        callback_url (str): URL that the review form will submit to (action) when
            the user is complete with their review
        callback_method (str): HTTP method the review form will submit to (method).
            Defaults to 'POST'

    Returns:
        template (html): json response with the detection web interface in html

    RESTful:
        Method: GET
        URL:    /api/review/detect/cnn/yolo/
    """
    ibs.web_check_uuids(image_uuid_list=[image_uuid])
    gid = ibs.get_image_gids_from_uuid(image_uuid)

    if gid is None:
        return 'INVALID IMAGE UUID'

    default_config = {
        'autointerest': False,
        'interest_bypass': False,
        'metadata': True,
        'metadata_viewpoint': False,
        'metadata_quality': False,
        'metadata_flags': True,
        'metadata_flags_aoi': True,
        'metadata_flags_multiple': False,
        'metadata_species': True,
        'metadata_label': True,
        'metadata_quickhelp': True,
        'parts': False,
        'modes_rectangle': True,
        'modes_diagonal': True,
        'modes_diagonal2': True,
        'staged': False,
    }

    if config is not None:
        default_config.update(config)

    gpath = ibs.get_image_thumbpath(gid, ensure_paths=True, draw_annots=False)
    image = ibs.get_images(gid)
    image_src = appf.embed_image_html(image)
    width, height = ibs.get_image_sizes(gid)

    if width <= 0 or width is None or height <= 0 or height is None:
        vals = (
            image_uuid,
            width,
            height,
        )
        raise IOError(
            'Image %r for review has either no width or no height (w = %s, h = %s)' % vals
        )

    annotation_list = []
    for result in result_list:
        quality = result.get('quality', None)
        if quality in [-1, None]:
            quality = 0
        elif quality <= 2:
            quality = 1
        elif quality > 2:
            quality = 2

        viewpoint1 = result.get('viewpoint1', None)
        viewpoint2 = result.get('viewpoint2', None)
        viewpoint3 = result.get('viewpoint3', None)

        if viewpoint1 is None and viewpoint2 is None and viewpoint3 is None:
            viewpoint = result.get('viewpoint', None)
            viewpoint1, viewpoint2, viewpoint3 = appf.convert_viewpoint_to_tuple(
                viewpoint
            )

        annotation_list.append(
            {
                'id': result.get('id', None),
                'left': 100.0 * (result.get('left', result['xtl']) / width),
                'top': 100.0 * (result.get('top', result['ytl']) / height),
                'width': 100.0 * (result['width'] / width),
                'height': 100.0 * (result['height'] / height),
                'species': result.get('species', result['class']),
                'theta': result.get('theta', 0.0),
                'viewpoint1': viewpoint1,
                'viewpoint2': viewpoint2,
                'viewpoint3': viewpoint3,
                'quality': quality,
                'multiple': 'true' if result.get('multiple', None) == 1 else 'false',
                'interest': 'true' if result.get('interest', None) == 1 else 'false',
            }
        )

    species = KEY_DEFAULTS[SPECIES_KEY]

    root_path = dirname(abspath(__file__))
    css_file_list = [
        ['include', 'jquery-ui', 'jquery-ui.min.css'],
        ['include', 'jquery.ui.rotatable', 'jquery.ui.rotatable.css'],
        ['css', 'style.css'],
    ]
    json_file_list = [
        ['include', 'jquery-ui', 'jquery-ui.min.js'],
        ['include', 'jquery.ui.rotatable', 'jquery.ui.rotatable.min.js'],
        ['include', 'bbox_annotator_percent.js'],
        ['javascript', 'script.js'],
        ['javascript', 'turk-detection.js'],
    ]

    if include_jquery:
        json_file_list = [['javascript', 'jquery.min.js']] + json_file_list

    EMBEDDED_CSS = ''
    EMBEDDED_JAVASCRIPT = ''

    css_template_fmtstr = '<style type="text/css" ia-dependency="css">%s</style>\n'
    json_template_fmtstr = (
        '<script type="text/javascript" ia-dependency="javascript">%s</script>\n'
    )
    for css_file in css_file_list:
        css_filepath_list = [root_path, 'static'] + css_file
        with open(join(*css_filepath_list)) as css_file:
            EMBEDDED_CSS += css_template_fmtstr % (css_file.read(),)

    for json_file in json_file_list:
        json_filepath_list = [root_path, 'static'] + json_file
        with open(join(*json_filepath_list)) as json_file:
            EMBEDDED_JAVASCRIPT += json_template_fmtstr % (json_file.read(),)

    species_rowids = ibs._get_all_species_rowids()
    species_nice_list = ibs.get_species_nice(species_rowids)

    combined_list = sorted(zip(species_nice_list, species_rowids))
    species_nice_list = [combined[0] for combined in combined_list]
    species_rowids = [combined[1] for combined in combined_list]

    species_text_list = ibs.get_species_texts(species_rowids)
    species_list = list(zip(species_nice_list, species_text_list))
    species_list = [('Unspecified', const.UNKNOWN)] + species_list

    # Collect mapping of species to parts
    aid_list = ibs.get_valid_aids()
    part_species_rowid_list = ibs.get_annot_species_rowids(aid_list)
    part_species_text_list = ibs.get_species_texts(part_species_rowid_list)
    part_rowids_list = ibs.get_annot_part_rowids(aid_list)
    part_types_list = map(ibs.get_part_types, part_rowids_list)

    zipped = list(zip(part_species_text_list, part_types_list))
    species_part_dict = {const.UNKNOWN: set([])}
    for part_species_text, part_type_list in zipped:
        if part_species_text not in species_part_dict:
            species_part_dict[part_species_text] = set([const.UNKNOWN])
        for part_type in part_type_list:
            species_part_dict[part_species_text].add(part_type)
            species_part_dict[const.UNKNOWN].add(part_type)
    # Add any images that did not get added because they aren't assigned any annotations
    for species_text in species_text_list:
        if species_text not in species_part_dict:
            species_part_dict[species_text] = set([const.UNKNOWN])
    for key in species_part_dict:
        species_part_dict[key] = sorted(list(species_part_dict[key]))
    species_part_dict_json = json.dumps(species_part_dict)

    orientation_flag = '0'
    if species is not None and 'zebra' in species:
        orientation_flag = '1'

    settings_key_list = [
        ('ia-detection-setting-orientation', orientation_flag),
        ('ia-detection-setting-parts-assignments', '1'),
        ('ia-detection-setting-toggle-annotations', '1'),
        ('ia-detection-setting-toggle-parts', '0'),
        ('ia-detection-setting-parts-show', '0'),
        ('ia-detection-setting-parts-hide', '0'),
    ]

    settings = {
        settings_key: request.cookies.get(settings_key, settings_default) == '1'
        for (settings_key, settings_default) in settings_key_list
    }

    return appf.template(
        'turk',
        'detection_insert',
        gid=gid,
        refer_aid=None,
        species=species,
        image_path=gpath,
        image_src=image_src,
        config=default_config,
        settings=settings,
        annotation_list=annotation_list,
        species_list=species_list,
        species_part_dict_json=species_part_dict_json,
        callback_url=callback_url,
        callback_method=callback_method,
        EMBEDDED_CSS=EMBEDDED_CSS,
        EMBEDDED_JAVASCRIPT=EMBEDDED_JAVASCRIPT,
    )


@register_api('/api/review/detect/cnn/yolo/', methods=['POST'])
def process_detection_html(ibs, **kwargs):
    """
    Process the return from the detection review interface.  Pass the POST result from the detection review form directly to this function unmodified.

    Returns:
        detection results (dict): Same format as `func:start_detect_image`

    RESTful:
        Method: POST
        URL:    /api/review/detect/cnn/yolo/
    """
    gid = int(request.form['detection-gid'])
    image_uuid = ibs.get_image_uuids(gid)
    width, height = ibs.get_image_sizes(gid)
    # Get aids
    annotation_list = json.loads(request.form['ia-detection-data'])

    viewpoint1_list = [
        int(annot['metadata'].get('viewpoint1', -1)) for annot in annotation_list
    ]
    viewpoint2_list = [
        int(annot['metadata'].get('viewpoint2', -1)) for annot in annotation_list
    ]
    viewpoint3_list = [
        int(annot['metadata'].get('viewpoint3', -1)) for annot in annotation_list
    ]
    zipped = list(zip(viewpoint1_list, viewpoint2_list, viewpoint3_list))
    viewpoint_list = [appf.convert_tuple_to_viewpoint(tup) for tup in zipped]

    result_list = [
        {
            'id': annot['label'],
            'xtl': int(width * (annot['percent']['left'] / 100.0)),
            'ytl': int(height * (annot['percent']['top'] / 100.0)),
            'left': int(width * (annot['percent']['left'] / 100.0)),
            'top': int(height * (annot['percent']['top'] / 100.0)),
            'width': int(width * (annot['percent']['width'] / 100.0)),
            'height': int(height * (annot['percent']['height'] / 100.0)),
            'theta': float(annot['angles']['theta']),
            'confidence': 1.0,
            'class': annot['label'],
            'species': annot['label'],
            'viewpoint': viewpoint,
            'quality': annot['metadata']['quality'],
            'multiple': annot['metadata']['multiple'],
            'interest': annot['highlighted'],
        }
        for annot, viewpoint in list(zip(annotation_list, viewpoint_list))
    ]
    result_dict = {
        'image_uuid_list': [image_uuid],
        'results_list': [result_list],
        'score_list': [1.0],
    }
    return result_dict


@register_ibs_method
@accessor_decors.getter_1to1
def detect_cnn_json(ibs, gid_list, detect_func, config={}, **kwargs):
    """
    Run animal detection in each image and returns json-ready formatted results, does not return annotations.

    Args:
        gid_list (list): list of image ids to run detection on

    Returns:
        results_dict (list): dict of detection results (not annotations)

    CommandLine:
        python -m wbia.web.apis_detect --test-detect_cnn_yolo_json

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.web.apis_detect import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb('testdb1')
        >>> gid_list = ibs.get_valid_gids()[0:2]
        >>> results_dict = ibs.detect_cnn_yolo_json(gid_list)
        >>> print(results_dict)
    """
    # TODO: Return confidence here as well
    image_uuid_list = ibs.get_image_uuids(gid_list)
    ibs.assert_valid_gids(gid_list)
    # Get detections from depc
    aids_list = detect_func(gid_list, **config)
    results_list = [
        [
            {
                'id': aid,
                'uuid': ibs.get_annot_uuids(aid),
                'xtl': ibs.get_annot_bboxes(aid)[0],
                'ytl': ibs.get_annot_bboxes(aid)[1],
                'left': ibs.get_annot_bboxes(aid)[0],
                'top': ibs.get_annot_bboxes(aid)[1],
                'width': ibs.get_annot_bboxes(aid)[2],
                'height': ibs.get_annot_bboxes(aid)[3],
                'theta': round(ibs.get_annot_thetas(aid), 4),
                'confidence': round(ibs.get_annot_detect_confidence(aid), 4),
                'class': ibs.get_annot_species_texts(aid),
                'species': ibs.get_annot_species_texts(aid),
                'viewpoint': ibs.get_annot_viewpoints(aid),
                'quality': ibs.get_annot_qualities(aid),
                'multiple': ibs.get_annot_multiple(aid),
                'interest': ibs.get_annot_interest(aid),
            }
            for aid in aid_list
        ]
        for aid_list in aids_list
    ]
    score_list = [0.0] * len(gid_list)
    # Wrap up results with other information
    results_dict = {
        'image_uuid_list': image_uuid_list,
        'results_list': results_list,
        'score_list': score_list,
    }
    return results_dict


@register_ibs_method
def detect_cnn_json_wrapper(ibs, image_uuid_list, detect_func, **kwargs):
    """
    Detect with CNN (general).

    REST:
        Method: GET
        URL: /api/detect/cnn/yolo/json/

    Args:
        image_uuid_list (list) : list of image uuids to detect on.
    """
    from wbia.web.apis_engine import ensure_uuid_list

    # Check UUIDs
    ibs.web_check_uuids(image_uuid_list=image_uuid_list)
    image_uuid_list = ensure_uuid_list(image_uuid_list)
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return detect_func(gid_list, **kwargs)


@register_ibs_method
@register_api('/api/detect/cnn/yolo/json/', methods=['POST'])
def detect_cnn_yolo_json_wrapper(ibs, image_uuid_list, **kwargs):
    return detect_cnn_json_wrapper(
        ibs, image_uuid_list, ibs.detect_cnn_yolo_json, **kwargs
    )


@register_ibs_method
@accessor_decors.getter_1to1
def detect_cnn_yolo_json(ibs, gid_list, config={}, **kwargs):
    return detect_cnn_json(ibs, gid_list, ibs.detect_cnn_yolo, config=config, **kwargs)


@register_ibs_method
@accessor_decors.getter_1toM
@register_api('/api/detect/cnn/yolo/', methods=['PUT', 'GET', 'POST'])
def detect_cnn_yolo(ibs, gid_list, model_tag=None, commit=True, testing=False, **kwargs):
    """
    Run animal detection in each image. Adds annotations to the database as they are found.

    Args:
        gid_list (list): list of image ids to run detection on

    Returns:
        aids_list (list): list of lists of annotation ids detected in each
            image

    CommandLine:
        python -m wbia.web.apis_detect --test-detect_cnn_yolo --show

    RESTful:
        Method: PUT, GET
        URL:    /api/detect/cnn/yolo/

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.web.apis_detect import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb('PZ_MTEST')
        >>> gid_list = ibs.get_valid_gids()[:5]
        >>> aids_list = ibs.detect_cnn_yolo(gid_list)
        >>> if ut.show_was_requested():
        >>>     import wbia.plottool as pt
        >>>     from wbia.viz import viz_image
        >>>     for fnum, gid in enumerate(gid_list):
        >>>         viz_image.show_image(ibs, gid, fnum=fnum)
        >>>     pt.show_if_requested()
        >>> # Remove newly detected annotations
        >>> ibs.delete_annots(ut.flatten(aids_list))
    """
    # TODO: Return confidence here as well
    depc = ibs.depc_image
    config = {
        'grid': False,
        'algo': 'yolo',
        'sensitivity': 0.2,
        'nms': True,
        'nms_thresh': 0.4,
    }
    if model_tag is not None:
        config['config_filepath'] = model_tag
        config['weight_filepath'] = model_tag

    config_str_list = ['config_filepath', 'weight_filepath'] + list(config.keys())
    for config_str in config_str_list:
        if config_str in kwargs:
            config[config_str] = kwargs[config_str]

    if testing:
        depc.delete_property('localizations', gid_list, config=config)

    results_list = depc.get_property('localizations', gid_list, None, config=config)

    if commit:
        aids_list = ibs.commit_localization_results(
            gid_list, results_list, note='cnnyolodetect', **kwargs
        )
        return aids_list
    else:
        return results_list

    # results_list = depc.get_property('detections', gid_list, None, config=config)
    # if commit:
    #     aids_list = ibs.commit_detection_results(gid_list, results_list, note='cnnyolodetect')
    #     return aids_list


@register_ibs_method
@register_api(
    '/api/models/cnn/lightnet/',
    methods=['PUT', 'GET', 'POST'],
    __api_plural_check__=False,
)
def models_cnn_lightnet(ibs, **kwargs):
    """
    Return the models (and their labels) for the YOLO CNN detector

    RESTful:
        Method: PUT, GET
        URL:    /api/labels/cnn/lightnet/
    """

    def identity(x):
        return x

    from wbia.algo.detect.lightnet import CONFIG_URL_DICT, _parse_class_list

    model_dict = ibs.models_cnn(CONFIG_URL_DICT, identity, _parse_class_list, **kwargs)
    return model_dict


@register_ibs_method
@register_api(
    '/api/models/cnn/yolo/', methods=['PUT', 'GET', 'POST'], __api_plural_check__=False
)
def models_cnn_yolo(ibs, **kwargs):
    """
    Return the models (and their labels) for the YOLO CNN detector

    RESTful:
        Method: PUT, GET
        URL:    /api/labels/cnn/yolo/
    """
    from pydarknet._pydarknet import (
        CONFIG_URL_DICT,
        _parse_classes_from_cfg,
        _parse_class_list,
    )

    model_dict = ibs.models_cnn(
        CONFIG_URL_DICT, _parse_classes_from_cfg, _parse_class_list, **kwargs
    )
    return model_dict


@register_ibs_method
def models_cnn(
    ibs,
    config_dict,
    parse_classes_func,
    parse_line_func,
    check_hash=False,
    hidden_models=[],
    **kwargs,
):
    import urllib

    model_dict = {}
    for config_tag in config_dict:
        if config_tag in hidden_models:
            continue

        try:
            config_url = config_dict[config_tag]
            classes_url = parse_classes_func(config_url)
            try:
                classes_filepath = ut.grab_file_url(
                    classes_url, appname='wbia', check_hash=check_hash
                )
                assert exists(classes_filepath)
            except (urllib.error.HTTPError, AssertionError):
                continue

            classes_filepath = ut.truepath(classes_filepath)
            line_list = parse_line_func(classes_filepath)
            model_dict[config_tag] = line_list
        except Exception:
            pass

    return model_dict


@register_ibs_method
@accessor_decors.getter_1toM
@register_api('/api/labeler/cnn/', methods=['PUT', 'GET', 'POST'])
def labeler_cnn(
    ibs, aid_list, testing=False, algo='pipeline', model_tag='candidacy', **kwargs
):
    depc = ibs.depc_annot
    config = {}

    if algo is not None:
        config['labeler_algo'] = algo
    if model_tag is not None:
        config['labeler_weight_filepath'] = model_tag

    if testing:
        depc.delete_property('labeler', aid_list, config=config)

    result_list = depc.get_property('labeler', aid_list, None, config=config)

    output_list = []
    for result in result_list:
        score, species, viewpoint, quality, orientation, probs = result
        output_list.append({'score': score, 'species': species, 'viewpoint': viewpoint})

    return output_list


@register_ibs_method
@accessor_decors.getter_1toM
@register_api('/api/aoi/cnn/', methods=['PUT', 'GET', 'POST'])
def aoi_cnn(ibs, aid_list, testing=False, model_tag='candidacy', **kwargs):
    depc = ibs.depc_annot
    config = {}

    if model_tag is not None:
        config['aoi_two_weight_filepath'] = model_tag

    if testing:
        depc.delete_property('aoi_two', aid_list, config=config)

    result_list = depc.get_property('aoi_two', aid_list, None, config=config)

    output_list = []
    for result in result_list:
        score, class_ = result
        output_list.append({'score': score, 'class': class_})

    return output_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/detect/cnn/yolo/exists/', methods=['GET'], __api_plural_check__=False)
def detect_cnn_yolo_exists(ibs, gid_list, testing=False):
    """
    Check to see if a detection has been completed.

    Args:
        gid_list (list): list of image ids to run detection on

    Returns:
        flag_list (list): list of flags for if the detection has been run on
            the image

    CommandLine:
        python -m wbia.web.apis_detect --test-detect_cnn_yolo_exists

    RESTful:
        Method: GET
        URL:    /api/detect/cnn/yolo/exists/

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.web.apis_detect import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb('PZ_MTEST')
        >>> gid_list = ibs.get_valid_gids()
        >>> depc = ibs.depc_image
        >>> aids_list = ibs.detect_cnn_yolo(gid_list[:3], testing=True)
        >>> result = ibs.detect_cnn_yolo_exists(gid_list[:5])
        >>> ibs.delete_annots(ut.flatten(aids_list))
        >>> print(result)
        [True, True, True, False, False]
    """
    depc = ibs.depc_image
    config = {
        'algo': 'yolo',
        'sensitivity': 0.2,
        'nms': True,
        'nms_thresh': 0.4,
    }
    score_list = depc.get_property(
        'localizations', gid_list, 'score', ensure=False, config=config
    )
    # score_list = depc.get_property('detections', gid_list, 'score', ensure=False, config=config)
    flag_list = [score is not None for score in score_list]
    return flag_list


@register_ibs_method
@register_api('/api/detect/cnn/lightnet/json/', methods=['POST'])
def detect_cnn_lightnet_json_wrapper(ibs, image_uuid_list, **kwargs):
    return detect_cnn_json_wrapper(
        ibs, image_uuid_list, ibs.detect_cnn_lightnet_json, **kwargs
    )


@register_ibs_method
@accessor_decors.getter_1to1
def detect_cnn_lightnet_json(ibs, gid_list, config={}, **kwargs):
    return detect_cnn_json(
        ibs, gid_list, ibs.detect_cnn_lightnet, config=config, **kwargs
    )


@register_ibs_method
@accessor_decors.getter_1toM
@register_api('/api/detect/cnn/lightnet/', methods=['PUT', 'GET', 'POST'])
def detect_cnn_lightnet(
    ibs, gid_list, model_tag=None, commit=True, testing=False, **kwargs
):
    """
    Run animal detection in each image. Adds annotations to the database as they are found.

    Args:
        gid_list (list): list of image ids to run detection on

    Returns:
        aids_list (list): list of lists of annotation ids detected in each
            image

    CommandLine:
        python -m wbia.web.apis_detect --test-detect_cnn_lightnet --show

    RESTful:
        Method: PUT, GET
        URL:    /api/detect/cnn/lightnet/

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.web.apis_detect import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb('PZ_MTEST')
        >>> gid_list = ibs.get_valid_gids()[:5]
        >>> aids_list = ibs.detect_cnn_lightnet(gid_list)
        >>> if ut.show_was_requested():
        >>>     import wbia.plottool as pt
        >>>     from wbia.viz import viz_image
        >>>     for fnum, gid in enumerate(gid_list):
        >>>         viz_image.show_image(ibs, gid, fnum=fnum)
        >>>     pt.show_if_requested()
        >>> # Remove newly detected annotations
        >>> ibs.delete_annots(ut.flatten(aids_list))
    """
    # TODO: Return confidence here as well
    depc = ibs.depc_image
    config = {
        'algo': 'lightnet',
        'sensitivity': 0.75,
        'nms': True,
        'nms_thresh': 0.4,
        'nms_aware': None,
    }

    if model_tag is not None:
        config['config_filepath'] = model_tag
        config['weight_filepath'] = model_tag

    config_str_list = ['config_filepath', 'weight_filepath'] + list(config.keys())
    for config_str in config_str_list:
        if config_str in kwargs:
            config[config_str] = kwargs[config_str]

    if testing:
        depc.delete_property('localizations', gid_list, config=config)

    results_list = depc.get_property('localizations', gid_list, None, config=config)

    if commit:
        aids_list = ibs.commit_localization_results(
            gid_list, results_list, note='cnnlightnetdetect', **kwargs
        )
        return aids_list
    else:
        return results_list


@register_ibs_method
def commit_localization_results(
    ibs,
    gid_list,
    results_list,
    note=None,
    labeler_algo='pipeline',
    labeler_model_tag=None,
    use_labeler_species=False,
    orienter_algo=None,
    orienter_model_tag=None,
    update_json_log=True,
    **kwargs,
):
    global_gid_list = []
    global_bbox_list = []
    global_theta_list = []
    global_class_list = []
    global_conf_list = []
    global_notes_list = []

    zipped_list = list(zip(gid_list, results_list))

    for gid, results in zipped_list:
        score, bbox_list, theta_list, conf_list, class_list = results
        num = len(bbox_list)
        gid_list_ = [gid] * num
        notes_list = [note] * num

        global_gid_list += list(gid_list_)
        global_bbox_list += list(bbox_list)
        global_theta_list += list(theta_list)
        global_class_list += list(class_list)
        global_conf_list += list(conf_list)
        global_notes_list += list(notes_list)

    assert len(global_gid_list) == len(global_bbox_list)
    assert len(global_gid_list) == len(global_theta_list)
    assert len(global_gid_list) == len(global_class_list)
    assert len(global_gid_list) == len(global_conf_list)
    assert len(global_gid_list) == len(global_notes_list)

    global_aid_list = ibs.add_annots(
        global_gid_list,
        global_bbox_list,
        global_theta_list,
        global_class_list,
        detect_confidence_list=global_conf_list,
        notes_list=global_notes_list,
        quiet_delete_thumbs=True,
        skip_cleaning=True,
    )

    global_aid_set = set(global_aid_list)
    aids_list = ibs.get_image_aids(gid_list)
    aids_list = [
        [aid for aid in aid_list_ if aid in global_aid_set] for aid_list_ in aids_list
    ]
    aid_list = ut.flatten(aids_list)

    if labeler_model_tag is not None:
        labeler_config = {}
        labeler_config['labeler_algo'] = labeler_algo
        labeler_config['labeler_weight_filepath'] = labeler_model_tag
        viewpoint_list = ibs.depc_annot.get_property(
            'labeler', aid_list, 'viewpoint', config=labeler_config
        )
        ibs.set_annot_viewpoints(aid_list, viewpoint_list)
        if use_labeler_species:
            species_list = ibs.depc_annot.get_property(
                'labeler', aid_list, 'species', config=labeler_config
            )
            ibs.set_annot_species(aid_list, species_list)

    if orienter_algo is not None:
        orienter_config = {}
        orienter_config['orienter_algo'] = orienter_algo
        orienter_config['orienter_weight_filepath'] = orienter_model_tag
        result_list = ibs.depc_annot.get_property(
            'orienter', aid_list, None, config=orienter_config
        )
        xtl_list = list(map(int, map(np.around, ut.take_column(result_list, 0))))
        ytl_list = list(map(int, map(np.around, ut.take_column(result_list, 1))))
        w_list = list(map(int, map(np.around, ut.take_column(result_list, 2))))
        h_list = list(map(int, map(np.around, ut.take_column(result_list, 3))))
        theta_list = ut.take_column(result_list, 4)
        bbox_list = list(zip(xtl_list, ytl_list, w_list, h_list))
        assert len(aid_list) == len(bbox_list)
        assert len(aid_list) == len(theta_list)
        if len(bbox_list) > 0:
            ibs.set_annot_bboxes(aid_list, bbox_list, theta_list=theta_list)

    ibs._clean_species()
    if update_json_log:
        ibs.log_detections(aid_list)

    return aids_list


@register_ibs_method
def commit_detection_results(
    ibs, gid_list, results_list, note=None, update_json_log=True
):
    zipped_list = list(zip(gid_list, results_list))
    aids_list = []
    for (
        gid,
        (score, bbox_list, theta_list, species_list, viewpoint_list, conf_list),
    ) in zipped_list:
        num = len(bbox_list)
        notes_list = None if note is None else [note] * num
        aid_list = ibs.add_annots(
            [gid] * num,
            bbox_list,
            theta_list,
            species_list,
            detect_confidence_list=conf_list,
            notes_list=notes_list,
            quiet_delete_thumbs=True,
            skip_cleaning=True,
        )
        ibs.set_annot_viewpoints(aid_list, viewpoint_list)
        # TODO ibs.set_annot_viewpoint_code(aid_list, viewpoint_list)
        aids_list.append(aid_list)
    ibs._clean_species()
    if update_json_log:
        aid_list = ut.flatten(aids_list)
        ibs.log_detections(aid_list)
    return aids_list


@register_ibs_method
def commit_detection_results_filtered(
    ibs,
    gid_list,
    filter_species_list=None,
    filter_viewpoint_list=None,
    note=None,
    update_json_log=True,
):
    depc = ibs.depc_image
    results_list = depc.get_property('detections', gid_list, None)
    zipped_list = list(zip(gid_list, results_list))
    aids_list = []
    for (
        gid,
        (score, bbox_list, theta_list, species_list, viewpoint_list, conf_list),
    ) in zipped_list:
        aid_list = []
        result_list = list(
            zip(bbox_list, theta_list, species_list, viewpoint_list, conf_list)
        )
        for bbox, theta, species, viewpoint, conf in result_list:
            if not (filter_species_list is None or species in filter_species_list):
                continue
            if not (filter_viewpoint_list is None or viewpoint in filter_viewpoint_list):
                continue
            note_ = None if note is None else [note]
            temp_list = ibs.add_annots(
                [gid],
                [bbox],
                [theta],
                [species],
                detect_confidence_list=[conf],
                notes_list=note_,
                quiet_delete_thumbs=True,
                skip_cleaning=True,
            )
            aid = temp_list[0]
            ibs.set_annot_viewpoints([aid], [viewpoint])
            # TODO ibs.set_annot_viewpoint_code([aid], [viewpoint])
            aid_list.append(aid)
        aids_list.append(aid_list)
    ibs._clean_species()
    if update_json_log:
        aid_list = ut.flatten(aids_list)
        ibs.log_detections(aid_list)
    return aids_list


@register_ibs_method
def log_detections(ibs, aid_list, fallback=True):
    import time
    import os

    json_log_path = ibs.get_logdir_local()
    json_log_filename = 'detections.json'
    json_log_filepath = os.path.join(json_log_path, json_log_filename)
    logger.info('Logging detections added to: %r' % (json_log_filepath,))

    try:
        # Log has never been made, create one
        if not os.path.exists(json_log_filepath):
            json_dict = {
                'updates': [],
            }
            json_str = ut.to_json(json_dict, pretty=True)
            with open(json_log_filepath, 'w') as json_log_file:
                json_log_file.write(json_str)
        # Get current log state
        with open(json_log_filepath, 'r') as json_log_file:
            json_str = json_log_file.read()
        json_dict = ut.from_json(json_str)
        # Get values
        db_name = ibs.get_db_name()
        db_init_uuid = ibs.get_db_init_uuid()
        # Zip all the updates together and write to updates list in dictionary
        gid_list = ibs.get_annot_gids(aid_list)
        bbox_list = ibs.get_annot_bboxes(aid_list)
        theta_list = ibs.get_annot_thetas(aid_list)
        zipped = list(zip(aid_list, gid_list, bbox_list, theta_list))
        for aid, gid, bbox, theta in zipped:
            json_dict['updates'].append(
                {
                    'time_unixtime': time.time(),
                    'db_name': db_name,
                    'db_init_uuid': db_init_uuid,
                    'image_rowid': gid,
                    'annot_rowid': aid,
                    'annot_bbox': bbox,
                    'annot_theta': theta,
                }
            )
        # Write new log state
        json_str = ut.to_json(json_dict, pretty=True)
        with open(json_log_filepath, 'w') as json_log_file:
            json_log_file.write(json_str)
    except Exception:
        if fallback:
            logger.info('WRITE DETECTION.JSON FAILED - ATTEMPTING FALLBACK')
            ut.delete(json_log_filepath)
            ibs.log_detections(aid_list, fallback=False)
        else:
            logger.info('WRITE DETECTION.JSON FAILED - FALLBACK FAILED')


@register_ibs_method
@register_api('/api/detect/species/enabled/', methods=['GET'], __api_plural_check__=False)
def has_species_detector(ibs, species_text):
    """
    TODO: extend to use non-constant species.

    RESTful:
        Method: GET
        URL:    /api/detect/species/enabled/
    """
    # FIXME: infer this
    return species_text in const.SPECIES_WITH_DETECTORS


@register_ibs_method
@register_api('/api/detect/species/', methods=['GET'], __api_plural_check__=False)
def get_species_with_detectors(ibs):
    """
    Get valid species for detection.

    RESTful:
        Method: GET
        URL:    /api/detect/species/
    """
    # FIXME: infer this
    return const.SPECIES_WITH_DETECTORS


@register_ibs_method
@register_api('/api/detect/species/working/', methods=['GET'], __api_plural_check__=False)
def get_working_species(ibs):
    """
    Get working species for detection.

    RESTful:
        Method: GET
        URL:    /api/detect/species/working/
    """
    RESTRICT_TO_ONLY_SPECIES_WITH_DETECTORS = ut.get_argflag('--no-allspecies')

    species_nice_list = ibs.get_all_species_nice()
    species_text_list = ibs.get_all_species_texts()
    species_tup_list = list(zip(species_nice_list, species_text_list))
    if RESTRICT_TO_ONLY_SPECIES_WITH_DETECTORS:
        working_species_tups = [
            species_tup
            for species_tup in species_tup_list
            if ibs.has_species_detector(species_tup[1])
        ]
    else:
        working_species_tups = species_tup_list
    return working_species_tups


@register_ibs_method
@register_api('/api/detect/whaleSharkInjury/', methods=['PUT', 'GET'])
def detect_ws_injury(ibs, gid_list):
    """
    Classify if a whale shark is injured.

    Args:
        gid_list (list): list of image ids to run classification on

    Returns:
        result_list (dictionary): predictions is list of strings representing a possible tag.
            confidences is a list of floats of correspoinding cofidence to the prediction

    """
    from wbia.scripts import labelShark

    labels = labelShark.classifyShark(ibs, gid_list)
    return labels



# -*- coding: utf-8 -*-
import logging
from wbia.algo.graph import demo
import utool as ut
from wbia.algo.graph.state import POSTV, NEGTV, INCMP, UNREV  # NOQA
from wbia.algo.graph.state import SAME, DIFF, NULL  # NOQA

print, rrr, profile = ut.inject2(__name__)
logger = logging.getLogger('wbia')


def do_infr_test(ccs, edges, new_edges):
    """
    Creates a graph with `ccs` + `edges` and then adds `new_edges`
    """
    # import networkx as nx
    import wbia.plottool as pt

    infr = demo.make_demo_infr(ccs, edges)

    if ut.show_was_requested():
        pt.qtensure()

    # Preshow
    fnum = 1
    if ut.show_was_requested():
        infr.set_node_attrs('shape', 'circle')
        infr.show(
            pnum=(2, 1, 1),
            fnum=fnum,
            show_unreviewed_edges=True,
            show_reviewed_cuts=True,
            splines='spline',
            show_inferred_diff=True,
            groupby='name_label',
            show_labels=True,
            pickable=True,
        )
        pt.set_title('pre-review')
        pt.gca().set_aspect('equal')
        infr.set_node_attrs('pin', 'true')
        # fig1 = pt.gcf()
        # fig1.canvas.mpl_connect('pick_event', ut.partial(on_pick, infr=infr))

    infr1 = infr
    infr2 = infr.copy()
    for new_edge in new_edges:
        aid1, aid2, data = new_edge
        evidence_decision = data['evidence_decision']
        infr2.add_feedback((aid1, aid2), evidence_decision)
    infr2.relabel_using_reviews(rectify=False)
    infr2.apply_nondynamic_update()

    # Postshow
    if ut.show_was_requested():
        infr2.show(
            pnum=(2, 1, 2),
            fnum=fnum,
            show_unreviewed_edges=True,
            show_inferred_diff=True,
            show_labels=True,
        )
        pt.gca().set_aspect('equal')
        pt.set_title('post-review')
        # fig2 = pt.gcf()
        # if fig2 is not fig1:
        #     fig2.canvas.mpl_connect('pick_event', ut.partial(on_pick, infr=infr2))

    class Checker(object):
        """
        Asserts pre and post test properties of the graph
        """

        def __init__(self, infr1, infr2):
            self._errors = []
            self.infr1 = infr1
            self.infr2 = infr2

        def __call__(self, infr, u, v, key, val, msg):
            data = infr.get_nonvisual_edge_data((u, v))
            if data is None:
                assert infr.graph.has_edge(u, v), 'uv=%r, %r does not exist' % (u, v)
            got = data.get(key)
            if got != val:
                msg1 = 'key=%s %r!=%r, ' % (key, got, val)
                errmsg = ''.join(
                    [
                        msg1,
                        msg,
                        '\nedge=',
                        ut.repr2((u, v)),
                        '\n',
                        infr.repr_edge_data(data),
                    ]
                )
                self._errors.append(errmsg)

        def custom_precheck(self, func):
            try:
                func(self.infr1)
            except AssertionError as ex:
                self._errors.append(str(ex))

        def after(self, errors=[]):
            """
            Delays error reporting until after visualization

            prints errors, then shows you the graph, then
            finally if any errors were discovered they are raised
            """

            errors = errors + self._errors
            if errors:
                ut.cprint('PRINTING %d FAILURE' % (len(errors)), 'red')
                for msg in errors:
                    logger.info(msg)
                ut.cprint('HAD %d FAILURE' % (len(errors)), 'red')
            if ut.show_was_requested():
                pt.all_figures_tile(percent_w=0.5)
                ut.show_if_requested()
            if errors:
                raise AssertionError('There were errors')

    check = Checker(infr1, infr2)
    return infr1, infr2, check


def case_negative_infr():
    """
    CommandLine:
        python -m wbia.algo.graph.tests.dyn_cases case_negative_infr --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.graph.tests.dyn_cases import *  # NOQA
        >>> case_negative_infr()
    """
    # Initial positive reviews
    ccs = [[1, 2, 3], [9]]
    # Add in initial reviews
    edges = [
        (9, 7, {'evidence_decision': NEGTV}),
        (1, 7, {'inferred_state': None}),
        (1, 9, {'inferred_state': None}),
    ]
    # Add in scored, but unreviewed edges
    new_edges = [(3, 9, {'evidence_decision': NEGTV})]
    infr1, infr2, check = do_infr_test(ccs, edges, new_edges)

    check(
        infr2,
        1,
        7,
        'inferred_state',
        None,
        'negative review of an edge should not jump more than one component',
    )

    check(
        infr2,
        1,
        9,
        'inferred_state',
        'diff',
        'negative review of an edge should cut within one jump',
    )

    check.after()


def case_match_infr():
    """
    CommandLine:
        python -m wbia.algo.graph.tests.dyn_cases case_match_infr --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.graph.tests.dyn_cases import *  # NOQA

        >>> case_match_infr()
    """
    # Initial positive reviews
    ccs = [[2], [7], [9, 10]]
    # Add in initial reviews
    edges = [
        (9, 8, {'evidence_decision': NEGTV}),
        (7, 2, {}),
    ]
    # Add in scored, but unreviewed edges
    edges += [
        (2, 8, {'inferred_state': None}),
        (2, 9, {'inferred_state': None}),
    ]
    new_edges = [(2, 10, {'evidence_decision': POSTV})]
    infr1, infr2, check = do_infr_test(ccs, edges, new_edges)

    # Checks out of cc inferences
    check(infr2, 2, 9, 'inferred_state', 'same', 'should infer a match')
    check(infr2, 2, 8, 'inferred_state', 'diff', 'should infer a negative')
    check(infr1, 2, 7, 'inferred_state', None, 'discon should have inference')

    check(infr2, 2, 7, 'inferred_state', None, 'discon should have inference')
    check.after()


def case_inconsistent():
    """
    CommandLine:
        python -m wbia.algo.graph.tests.dyn_cases case_inconsistent --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.graph.tests.dyn_cases import *  # NOQA
        >>> case_inconsistent()
    """
    ccs = [[1, 2], [3, 4, 5]]  # [6, 7]]
    edges = [
        (2, 3, {'evidence_decision': NEGTV}),
    ]
    edges += [
        (4, 1, {'inferred_state': None}),
        # (2, 7, {'inferred_state': None}),
    ]
    new_edges = [(1, 5, {'evidence_decision': POSTV})]
    infr1, infr2, check = do_infr_test(ccs, edges, new_edges)
    # Make sure the previously inferred edge is no longer inferred
    check(infr1, 4, 1, 'inferred_state', 'diff', 'should initially be an inferred diff')
    check(
        infr2,
        4,
        1,
        'inferred_state',
        'inconsistent_internal',
        'should not be inferred after incon',
    )
    check(infr2, 4, 3, 'maybe_error', True, 'need to have a maybe split')
    check.after()


def case_redo_incon():
    """
    CommandLine:
        python -m wbia.algo.graph.tests.dyn_cases case_redo_incon --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.graph.tests.dyn_cases import *  # NOQA
        >>> case_redo_incon()
    """
    ccs = [[1, 2], [3, 4]]  # [6, 7]]
    edges = [
        (2, 3, {'evidence_decision': NEGTV}),
        (1, 4, {'evidence_decision': NEGTV}),
    ]
    edges += []
    new_edges = [(2, 3, {'evidence_decision': POSTV})]
    infr1, infr2, check = do_infr_test(ccs, edges, new_edges)

    maybe_splits = infr2.get_edge_attrs('maybe_error', default=None)
    logger.info('maybe_splits = %r' % (maybe_splits,))
    if not any(maybe_splits.values()):
        ut.cprint('FAILURE', 'red')
        logger.info('At least one edge should be marked as a split')

    check.after()


def case_override_inference():
    """
    CommandLine:
        python -m wbia.algo.graph.tests.dyn_cases case_override_inference --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.graph.tests.dyn_cases import *  # NOQA
        >>> case_override_inference()
    """
    ccs = [[1, 2, 3, 4, 5]]
    edges = [
        (1, 3, {'inferred_state': 'same'}),
        (1, 4, {'inferred_state': 'same'}),
        # (1, 5, {'inferred_state': 'same'}),
        (1, 2, {'inferred_state': 'same', 'num_reviews': 2}),
        (2, 3, {'inferred_state': 'same', 'num_reviews': 2}),
        (2, 4, {'inferred_state': 'same'}),
        (2, 5, {'inferred_state': 'same'}),
        (3, 4, {'inferred_state': 'same', 'num_reviews': 100}),
        (4, 5, {'inferred_state': 'same', 'num_reviews': 0.01}),
    ]
    edges += []
    new_edges = [
        (1, 5, {'evidence_decision': NEGTV}),
        (5, 2, {'evidence_decision': NEGTV}),
    ]
    infr1, infr2, check = do_infr_test(ccs, edges, new_edges)
    # Make sure that the inferred edges are no longer inferred when an
    # inconsistent case is introduced
    check(infr2, 1, 4, 'maybe_error', None, 'should not split inferred edge')
    check(infr2, 4, 5, 'maybe_error', True, 'split me')
    check(
        infr2,
        5,
        2,
        'inferred_state',
        'inconsistent_internal',
        'inference should be overriden',
    )
    check.after()


def case_undo_match():
    """
    CommandLine:
        python -m wbia.algo.graph.tests.dyn_cases case_undo_match --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.graph.tests.dyn_cases import *  # NOQA
        >>> case_undo_match()
    """
    ccs = [[1, 2]]
    edges = []
    new_edges = [(1, 2, {'evidence_decision': NEGTV})]
    infr1, infr2, check = do_infr_test(ccs, edges, new_edges)

    check(infr2, 1, 2, 'inferred_state', 'diff', 'should have cut edge')
    check.after()


def case_undo_negative():
    """
    CommandLine:
        python -m wbia.algo.graph.tests.dyn_cases case_undo_negative --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.graph.tests.dyn_cases import *  # NOQA
        >>> case_undo_negative()
    """
    ccs = [[1], [2]]
    edges = [
        (1, 2, {'evidence_decision': NEGTV}),
    ]
    new_edges = [(1, 2, {'evidence_decision': POSTV})]
    infr1, infr2, check = do_infr_test(ccs, edges, new_edges)
    check(infr2, 1, 2, 'inferred_state', 'same', 'should have matched edge')
    check.after()


def case_incon_removes_inference():
    """
    CommandLine:
        python -m wbia.algo.graph.tests.dyn_cases case_incon_removes_inference --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.graph.tests.dyn_cases import *  # NOQA
        >>> case_incon_removes_inference()
    """
    ccs = [[1, 2, 3], [4, 5, 6]]
    edges = [
        (3, 4, {'evidence_decision': NEGTV}),
        (1, 5, {'evidence_decision': NEGTV}),
        (2, 5, {}),
        (1, 6, {}),
    ]
    new_edges = [(3, 4, {'evidence_decision': POSTV})]
    infr1, infr2, check = do_infr_test(ccs, edges, new_edges)

    check(infr1, 2, 5, 'inferred_state', 'diff', 'should be preinferred')
    check(
        infr2,
        2,
        5,
        'inferred_state',
        'inconsistent_internal',
        'should be uninferred on incon',
    )
    check.after()


def case_inferable_notcomp1():
    """
    make sure notcomparable edges can be inferred

    CommandLine:
        python -m wbia.algo.graph.tests.dyn_cases case_inferable_notcomp1 --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.graph.tests.dyn_cases import *  # NOQA
        >>> case_inferable_notcomp1()
    """
    ccs = [[1, 2], [3, 4]]
    edges = [
        (2, 3, {'evidence_decision': NEGTV}),
    ]
    new_edges = [(1, 4, {'evidence_decision': INCMP})]
    infr1, infr2, check = do_infr_test(ccs, edges, new_edges)
    check(infr2, 1, 4, 'inferred_state', 'diff', 'should be inferred')
    check.after()


def case_inferable_update_notcomp():
    """
    make sure inference updates for nocomparable edges

    CommandLine:
        python -m wbia.algo.graph.tests.dyn_cases case_inferable_update_notcomp --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.graph.tests.dyn_cases import *  # NOQA
        >>> case_inferable_update_notcomp()
    """
    ccs = [[1, 2], [3, 4]]
    edges = [
        (2, 3, {'evidence_decision': NEGTV}),
        (1, 4, {'evidence_decision': INCMP}),
    ]
    new_edges = [(2, 3, {'evidence_decision': POSTV})]
    infr1, infr2, check = do_infr_test(ccs, edges, new_edges)
    check(infr1, 1, 4, 'inferred_state', 'diff', 'should be inferred diff')
    check(infr2, 1, 4, 'inferred_state', 'same', 'should be inferred same')
    check.after()


def case_notcomp_remove_infr():
    """
    CommandLine:
        python -m wbia.algo.graph.tests.dyn_cases case_notcomp_remove_infr --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.graph.tests.dyn_cases import *  # NOQA
        >>> case_notcomp_remove_infr()
    """
    ccs = [[1, 2, 3], [4, 5, 6]]
    edges = [
        (1, 4, {'evidence_decision': POSTV}),
        # (1, 4, {'evidence_decision': INCMP}),
        (2, 5, {'evidence_decision': INCMP}),
        (3, 6, {'evidence_decision': INCMP}),
    ]
    new_edges = [(1, 4, {'evidence_decision': INCMP})]
    infr1, infr2, check = do_infr_test(ccs, edges, new_edges)
    check(infr2, 1, 4, 'inferred_state', INCMP, 'can not infer match here!')
    check(infr2, 2, 5, 'inferred_state', INCMP, 'can not infer match here!')
    check(infr2, 3, 6, 'inferred_state', INCMP, 'can not infer match here!')
    check.after()


def case_notcomp_remove_cuts():
    """
    CommandLine:
        python -m wbia.algo.graph.tests.dyn_cases case_notcomp_remove_cuts --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.graph.tests.dyn_cases import *  # NOQA
        >>> case_notcomp_remove_cuts()
    """
    ccs = [[1, 2, 3], [4, 5, 6]]
    edges = [
        (1, 4, {'evidence_decision': NEGTV}),
        # (1, 4, {'evidence_decision': INCMP}),
        (2, 5, {'evidence_decision': INCMP}),
        (3, 6, {'evidence_decision': INCMP}),
    ]
    new_edges = [(1, 4, {'evidence_decision': INCMP})]
    infr1, infr2, check = do_infr_test(ccs, edges, new_edges)
    check(infr1, 1, 4, 'inferred_state', 'diff', 'should infer diff!')
    check(infr1, 2, 5, 'inferred_state', 'diff', 'should infer diff!')
    check(infr1, 3, 6, 'inferred_state', 'diff', 'should infer diff!')
    check(infr2, 1, 4, 'evidence_decision', INCMP, 'can not infer cut here!')
    check(infr2, 2, 5, 'inferred_state', INCMP, 'can not infer cut here!')
    check(infr2, 3, 6, 'inferred_state', INCMP, 'can not infer cut here!')
    check.after()


def case_keep_in_cc_infr_post_negative():
    """
    CommandLine:
        python -m wbia.algo.graph.tests.dyn_cases case_keep_in_cc_infr_post_negative --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.graph.tests.dyn_cases import *  # NOQA
        >>> case_keep_in_cc_infr_post_negative()
    """
    ccs = [[1, 2, 3], [4]]
    edges = [(1, 3), (1, 4), (2, 4), (3, 4)]
    new_edges = [(4, 2, {'evidence_decision': NEGTV})]
    infr1, infr2, check = do_infr_test(ccs, edges, new_edges)
    check(infr1, 3, 4, 'inferred_state', None, 'should be no inference')
    check(infr1, 1, 3, 'inferred_state', 'same', 'should be inferred')
    check(infr2, 1, 3, 'inferred_state', 'same', 'should remain inferred')
    check(infr2, 3, 4, 'inferred_state', 'diff', 'should become inferred')
    check.after()


def case_keep_in_cc_infr_post_notcomp():
    """
    CommandLine:
        python -m wbia.algo.graph.tests.dyn_cases case_keep_in_cc_infr_post_notcomp --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.graph.tests.dyn_cases import *  # NOQA
        >>> case_keep_in_cc_infr_post_notcomp()
    """
    ccs = [[1, 2, 3], [4]]
    edges = [(1, 3), (1, 4), (2, 4), (3, 4)]
    new_edges = [(4, 2, {'evidence_decision': INCMP})]
    infr1, infr2, check = do_infr_test(ccs, edges, new_edges)
    check(infr1, 3, 4, 'inferred_state', None, 'should not be inferred')
    check(infr1, 1, 3, 'inferred_state', 'same', 'should be inferred')
    check(infr2, 1, 3, 'inferred_state', 'same', 'should remain inferred')
    check(infr2, 3, 4, 'inferred_state', None, 'should not become inferred')
    check.after()


def case_out_of_subgraph_modification():
    """
    CommandLine:
        python -m wbia.algo.graph.tests.dyn_cases case_out_of_subgraph_modification --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.graph.tests.dyn_cases import *  # NOQA
        >>> case_out_of_subgraph_modification()
    """
    # A case where a review between two ccs modifies state outside of
    # the subgraph of ccs
    ccs = [[1, 2], [3, 4], [5, 6]]
    edges = [(2, 6), (4, 5, {'evidence_decision': NEGTV})]
    new_edges = [(2, 3, {'evidence_decision': POSTV})]
    infr1, infr2, check = do_infr_test(ccs, edges, new_edges)
    check(infr1, 2, 6, 'inferred_state', None, 'should not be inferred')
    check(infr2, 2, 6, 'inferred_state', 'diff', 'should be inferred')
    check.after()


def case_flag_merge():
    """
    CommandLine:
        python -m wbia.algo.graph.tests.dyn_cases case_flag_merge --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.graph.tests.dyn_cases import *  # NOQA
        >>> case_flag_merge()
    """
    # A case where a review between two ccs modifies state outside of
    # the subgraph of ccs
    ccs = []
    edges = [
        (1, 2, {'evidence_decision': POSTV, 'num_reviews': 2}),
        (4, 1, {'evidence_decision': POSTV, 'num_reviews': 1}),
        (2, 4, {'evidence_decision': NEGTV, 'num_reviews': 1}),
    ]
    # Ensure that the negative edge comes back as potentially in error
    new_edges = [(1, 4, {'evidence_decision': POSTV})]
    infr1, infr2, check = do_infr_test(ccs, edges, new_edges)

    @check.custom_precheck
    def check_pre_state(infr):
        assert infr.nid_to_errors[1] == {(1, 4)}

    check(infr1, 2, 4, 'maybe_error', None, 'match edge should flag first None')
    check(infr1, 1, 4, 'maybe_error', True, 'match edge should flag first True')
    check(infr2, 2, 4, 'maybe_error', True, 'negative edge should flag second True')
    check(infr2, 1, 4, 'maybe_error', None, 'negative edge should flag second None')
    check.after()


def case_all_types():
    """
    CommandLine:
        python -m wbia.algo.graph.tests.dyn_cases case_all_types --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.algo.graph.tests.dyn_cases import *  # NOQA
        >>> case_all_types()
    """
    # A case where a review between two ccs modifies state outside of
    # the subgraph of ccs
    ccs = []
    # Define edges within components
    edges = [
        # Inconsistent component
        (11, 12, {'evidence_decision': POSTV}),
        (12, 13, {'evidence_decision': POSTV}),
        (11, 13, {'evidence_decision': NEGTV}),
        (11, 14, {'evidence_decision': POSTV}),
        (12, 14, {'evidence_decision': POSTV}),
        (13, 14, {}),
        (11, 15, {'evidence_decision': POSTV}),
        (12, 15, {'evidence_decision': INCMP}),
        # Positive component (with notcomp)
        (21, 22, {'evidence_decision': POSTV}),
        (22, 23, {'evidence_decision': POSTV}),
        (21, 23, {'evidence_decision': INCMP}),
        (21, 24, {'evidence_decision': POSTV}),
        (22, 24, {'evidence_decision': POSTV}),
        (23, 24, {}),
        # Positive component (with unreview)
        (31, 32, {'evidence_decision': POSTV}),
        (32, 33, {'evidence_decision': POSTV}),
        (31, 33, {'evidence_decision': POSTV}),
        (31, 34, {'evidence_decision': POSTV}),
        (32, 34, {'evidence_decision': POSTV}),
        (33, 34, {}),
        # Positive component
        (41, 42, {'evidence_decision': POSTV}),
        (42, 43, {'evidence_decision': POSTV}),
        (41, 43, {'evidence_decision': POSTV}),
        # Positive component (extra)
        (51, 52, {'evidence_decision': POSTV}),
        (52, 53, {'evidence_decision': POSTV}),
        (51, 53, {'evidence_decision': POSTV}),
        # Positive component (isolated)
        (61, 62, {'evidence_decision': POSTV}),
    ]
    # Define edges between components
    edges += [
        # 1 - 2
        (11, 21, {}),
        (12, 22, {}),
        # 1 - 3
        (11, 31, {}),
        (12, 32, {'evidence_decision': NEGTV}),
        (13, 33, {}),
        # 1 - 4
        (11, 41, {}),
        (12, 42, {'evidence_decision': INCMP}),
        (13, 43, {}),
        # 1 - 5
        (11, 51, {'evidence_decision': INCMP}),
        (12, 52, {'evidence_decision': NEGTV}),
        (13, 53, {}),
        # 2 - 3
        (21, 31, {'evidence_decision': INCMP}),
        (22, 32, {}),
        # 2 - 4
        (21, 41, {}),
        (22, 42, {}),
        # 2 - 5
        (21, 51, {'evidence_decision': INCMP}),
        (22, 52, {'evidence_decision': NEGTV}),
        # 3 - 4
        (31, 41, {'evidence_decision': NEGTV}),
        (32, 42, {}),
    ]
    # Ensure that the negative edge comes back as potentially in error
    # new_edges = [(2, 5, {'evidence_decision': POSTV})]
    new_edges = []
    infr1, infr2, check = do_infr_test(ccs, edges, new_edges)
    errors = []
    for u, v, d in infr2.graph.edges(data=True):
        state = d.get('inferred_state', '')
        if u < 20 or v < 20:
            if state is not None and 'inconsistent' not in state:
                logger.info('u, v, state = %r, %r, %r' % (u, v, state))
                err = AssertionError('all of cc0 should be incon')
                logger.info(err)
                errors.append(err)
        else:
            if state is not None and 'inconsistent' in state:
                logger.info('u, v, state = %r, %r, %r' % (u, v, state))
                err = AssertionError('outside of cc0 should not be incon')
                logger.info(err)
                errors.append(err)
    check(
        infr1,
        13,
        14,
        'inferred_state',
        'inconsistent_internal',
        'notcomp edge should be incon',
    )
    check(infr1, 21, 31, 'inferred_state', INCMP, 'notcomp edge should remain notcomp')
    check(infr1, 22, 32, 'inferred_state', None, 'notcomp edge should transfer knowledge')
    check(
        infr1,
        12,
        42,
        'inferred_state',
        'inconsistent_external',
        'inconsistency should override notcomp',
    )
    # check(infr1, 1, 4, 'maybe_error', True, 'match edge should flag first')
    # check(infr2, 2, 4, 'maybe_error', True, 'negative edge should flag second')
    # check(infr2, 1, 4, 'maybe_error', False, 'negative edge should flag second')
    check.after(errors)



# -*- coding: utf-8 -*-
import logging
import six
from wbia import constants as const
from wbia.control import accessor_decors, controller_inject
from wbia.control.controller_inject import make_ibs_register_decorator
import functools
import utool as ut
import uuid

print, rrr, profile = ut.inject2(__name__)
logger = logging.getLogger('wbia')


IMAGESET_OCCURRENCE_FLAG = 'imageset_occurrence_flag'
IMAGESET_END_TIME_POSIX = 'imageset_end_time_posix'
IMAGESET_GPS_LAT = 'imageset_gps_lat'
IMAGESET_GPS_LON = 'imageset_gps_lon'
IMAGESET_NOTE = 'imageset_note'
IMAGESET_PROCESSED_FLAG = 'imageset_processed_flag'
IMAGESET_ROWID = 'imageset_rowid'
IMAGESET_SHIPPED_FLAG = 'imageset_shipped_flag'
IMAGESET_START_TIME_POSIX = 'imageset_start_time_posix'
IMAGESET_SMART_WAYPOINT_ID = 'imageset_smart_waypoint_id'
IMAGESET_SMART_XML_FNAME = 'imageset_smart_xml_fname'


CLASS_INJECT_KEY, register_ibs_method = make_ibs_register_decorator(__name__)


register_api = controller_inject.get_wbia_flask_api(__name__)


@register_ibs_method
@accessor_decors.ider
def _get_all_imageset_rowids(ibs):
    r"""
    Returns:
        list_ (list):  all unfiltered imgsetids (imageset rowids)
    """
    all_imgsetids = ibs.db.get_all_rowids(const.IMAGESET_TABLE)
    return all_imgsetids


@register_ibs_method
@accessor_decors.ider
def _get_all_imgsetids(ibs):
    r"""
    alias
    """
    return _get_all_imageset_rowids(ibs)


@register_ibs_method
@accessor_decors.ider
@register_api('/api/imageset/', methods=['GET'])
def get_valid_imgsetids(
    ibs,
    min_num_gids=0,
    processed=None,
    shipped=None,
    is_occurrence=None,
    is_special=None,
):
    r"""
    FIX NAME imgagesetids

    Returns:
        list_ (list):  list of all imageset ids

    RESTful:
        Method: GET
        URL:    /api/imageset/
    """
    imgsetid_list = ibs._get_all_imgsetids()
    if min_num_gids > 0:
        num_gids_list = ibs.get_imageset_num_gids(imgsetid_list)
        flag_list = [num_gids >= min_num_gids for num_gids in num_gids_list]
        imgsetid_list = ut.compress(imgsetid_list, flag_list)
    if processed is not None:
        flag_list = ibs.get_imageset_processed_flags(imgsetid_list)
        isvalid_list = [flag == 1 if processed else flag == 0 for flag in flag_list]
        imgsetid_list = ut.compress(imgsetid_list, isvalid_list)
    if shipped is not None:
        flag_list = ibs.get_imageset_shipped_flags(imgsetid_list)
        isvalid_list = [flag == 1 if shipped else flag == 0 for flag in flag_list]
        imgsetid_list = ut.compress(imgsetid_list, isvalid_list)
    if is_occurrence is not None:
        flag_list = ibs.get_imageset_occurrence_flags(imgsetid_list)
        isvalid_list = [flag == is_occurrence for flag in flag_list]
        imgsetid_list = ut.compress(imgsetid_list, isvalid_list)
    if is_special is not None:
        flag_list = ibs.is_special_imageset(imgsetid_list)
        isvalid_list = [flag == is_special for flag in flag_list]
        imgsetid_list = ut.compress(imgsetid_list, isvalid_list)
    return imgsetid_list


@register_ibs_method
def is_special_imageset(ibs, imgsetid_list):
    imagesettext_list = ibs.get_imageset_text(imgsetid_list)
    isspecial_list = [
        str(imagesettext) in set(const.SPECIAL_IMAGESET_LABELS)
        for imagesettext in imagesettext_list
    ]
    return isspecial_list


@register_ibs_method
@accessor_decors.adder
@register_api('/api/imageset/', methods=['POST'])
def add_imagesets(
    ibs,
    imagesettext_list,
    imageset_uuid_list=None,
    notes_list=None,
    occurence_flag_list=None,
):
    r"""
    Adds a list of imagesets.

    Args:
        imagesettext_list (list):
        imageset_uuid_list (list):
        notes_list (list):

    Returns:
        imgsetid_list (list): added imageset rowids

    RESTful:
        Method: POST
        URL:    /api/imageset/
    """
    if ut.VERBOSE:
        logger.info('[ibs] adding %d imagesets' % len(imagesettext_list))
    # Add imageset text names to database
    if notes_list is None:
        notes_list = [''] * len(imagesettext_list)
    if imageset_uuid_list is None:
        imageset_uuid_list = [uuid.uuid4() for _ in range(len(imagesettext_list))]
    if occurence_flag_list is None:
        occurence_flag_list = [0] * len(imagesettext_list)
    colnames = [
        'imageset_text',
        'imageset_uuid',
        'imageset_occurrence_flag',
        'imageset_note',
    ]
    params_iter = zip(
        imagesettext_list, imageset_uuid_list, occurence_flag_list, notes_list
    )
    get_rowid_from_superkey = functools.partial(
        ibs.get_imageset_imgsetids_from_text, ensure=False
    )
    imgsetid_list = ibs.db.add_cleanly(
        const.IMAGESET_TABLE, colnames, params_iter, get_rowid_from_superkey
    )
    return imgsetid_list


# SETTERS::IMAGESET


@register_ibs_method
@accessor_decors.setter
@register_api('/api/imageset/text/', methods=['PUT'])
def set_imageset_text(ibs, imgsetid_list, imageset_text_list):
    r"""
    Sets names of imagesets (groups of animals)

    RESTful:
        Method: PUT
        URL:    /api/imageset/text/
    """
    # Special set checks
    if any(ibs.is_special_imageset(imgsetid_list)):
        raise ValueError('cannot rename special imagesets')
    id_iter = ((imgsetid,) for imgsetid in imgsetid_list)
    val_list = ((imageset_text,) for imageset_text in imageset_text_list)
    ibs.db.set(const.IMAGESET_TABLE, ('imageset_text',), val_list, id_iter)


#
# GETTERS::IMAGESET


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/imageset/num/image/reviewed/', methods=['GET'])
def get_imageset_num_imgs_reviewed(ibs, imgsetid_list):
    r"""
    RESTful:
        Method: GET
        URL:    /api/imageset/num/image/reviewed/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_imageset_funcs import *  # NOQA
        >>> ibs, config2_ = testdata_ibs()
        >>> # Reset and compute imagesets
        >>> ibs.delete_all_imagesets()
        >>> ibs.compute_occurrences(config={'use_gps': False, 'seconds_thresh': 600})
        >>> imgsetid_list = ibs.get_valid_imgsetids()
        >>> num_reviwed_list = ibs.get_imageset_num_imgs_reviewed(imgsetid_list)
        >>> result = num_reviwed_list
        >>> print(result)
        [0, 0]
    """
    gids_list = ibs.get_imageset_gids(imgsetid_list)
    flags_list = ibs.unflat_map(ibs.get_image_reviewed, gids_list)
    num_reviwed_list = [sum(flags) for flags in flags_list]
    return num_reviwed_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/imageset/num/annot/reviewed/', methods=['GET'])
def get_imageset_num_annots_reviewed(ibs, imgsetid_list):
    r"""
    RESTful:
        Method: GET
        URL:    /api/imageset/num/annot/reviewed/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_imageset_funcs import *  # NOQA
        >>> ibs, config2_ = testdata_ibs()
        >>> # Reset and compute imagesets
        >>> ibs.delete_all_imagesets()
        >>> ibs.compute_occurrences(config={'use_gps': False, 'seconds_thresh': 600})
        >>> imgsetid_list = ibs.get_valid_imgsetids()
        >>> num_reviwed_list = ibs.get_imageset_num_imgs_reviewed(imgsetid_list)
        >>> result = num_reviwed_list
        >>> print(result)
        [0, 0]
    """
    aids_list = ibs.get_imageset_aids(imgsetid_list)
    flags_list = ibs.unflat_map(ibs.get_annot_reviewed, aids_list)
    num_reviwed_list = [sum(flags) for flags in flags_list]
    return num_reviwed_list


@register_ibs_method
@accessor_decors.getter_1to1
# @register_api('/api/imageset/num/annotmatch/reviewed/', methods=['GET'])
def get_imageset_num_annotmatch_reviewed(ibs, imgsetid_list):
    r"""
    RESTful:
        Method: GET
        URL:    /api/imageset/num/annotmatch/reviewed/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_imageset_funcs import *  # NOQA
        >>> import wbia  # NOQA
        >>> ibs = wbia.opendb('testdb1')
        >>> imgsetid_list = ibs._get_all_imageset_rowids()
        >>> num_annots_reviewed_list = ibs.get_imageset_num_annotmatch_reviewed(imgsetid_list)
    """
    aids_list = ibs.get_imageset_custom_filtered_aids(imgsetid_list)
    has_revieweds_list = ibs.unflat_map(
        ibs.get_annot_has_reviewed_matching_aids, aids_list
    )
    num_annots_reviewed_list = list(map(sum, has_revieweds_list))
    return num_annots_reviewed_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/imageset/num/name/exemplar/', methods=['GET'])
def get_imageset_num_names_with_exemplar(ibs, imgsetid_list):
    r"""
    RESTful:
        Method: GET
        URL:    /api/imageset/num/name/exemplar/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_imageset_funcs import *  # NOQA
        >>> import wbia  # NOQA
        >>> ibs = wbia.opendb('testdb1')
        >>> imgsetid_list = ibs._get_all_imageset_rowids()
        >>> num_annots_reviewed_list = ibs.get_imageset_num_annotmatch_reviewed(imgsetid_list)
    """
    aids_list = ibs.get_imageset_custom_filtered_aids(imgsetid_list)
    exflags_list = ibs.unflat_map(ibs.get_annot_exemplar_flags, aids_list)
    nids_list = ibs.unflat_map(ibs.get_annot_name_rowids, aids_list)
    groups_list = [
        ut.group_items(exflags, nids) for exflags, nids in zip(exflags_list, nids_list)
    ]
    # num_names_list = [len(groups) for groups in groups_list]
    num_exemplared_names_list = [
        sum([any(exflags) for exflags in six.itervalues(groups)])
        for groups in groups_list
    ]
    return num_exemplared_names_list


@register_ibs_method
@accessor_decors.getter_1to1
def get_imageset_fraction_names_with_exemplar(ibs, imgsetid_list):
    r"""
    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.control.manual_imageset_funcs import *  # NOQA
        >>> import wbia  # NOQA
        >>> ibs = wbia.opendb('testdb2')
        >>> imgsetid_list = ibs._get_all_imageset_rowids()
        >>> fraction_exemplared_names_list = ibs.get_imageset_fraction_names_with_exemplar(imgsetid_list)
    """
    aids_list = ibs.get_imageset_custom_filtered_aids(imgsetid_list)
    # exflags_list = ibs.unflat_map(ibs.get_annot_exemplar_flags, aids_list)
    nids_list = list(
        map(list, map(set, ibs.unflat_map(ibs.get_annot_name_rowids, aids_list)))
    )
    aids_list_list = ibs.unflat_map(ibs.get_name_aids, nids_list)
    flags_list_list = list(
        map(lambda x: ibs.unflat_map(ibs.get_annot_exemplar_flags, x), aids_list_list)
    )
    # groups_list = [ut.group_items(exflags, nids)
    #               for exflags, nids in zip(exflags_list, nids_list)]
    num_names_list = list(map(len, nids_list))
    num_exemplared_names_list = [
        sum([any(exflags) for exflags in flags_list]) for flags_list in flags_list_list
    ]
    fraction_exemplared_names_list = [
        None if num_names == 0 else num_exemplared_names / num_names
        for num_exemplared_names, num_names in zip(
            num_exemplared_names_list, num_names_list
        )
    ]
    return fraction_exemplared_names_list


@register_ibs_method
@accessor_decors.getter_1to1
def get_imageset_fraction_annotmatch_reviewed(ibs, imgsetid_list):
    aids_list = ibs.get_imageset_custom_filtered_aids(imgsetid_list)
    flags_list = ibs.unflat_map(ibs.get_annot_has_reviewed_matching_aids, aids_list)
    fraction_annotmatch_reviewed_list = [
        None if len(flags) == 0 else sum(flags) / len(flags) for flags in flags_list
    ]
    return fraction_annotmatch_reviewed_list


@register_ibs_method
# @register_api('/api/imageset/aids/filtered/custom/', methods=['GET'])
def get_imageset_custom_filtered_aids(ibs, imgsetid_list):
    r"""
    hacks to filter aids to only certain views and qualities
    """
    aids_list_ = ibs.get_imageset_aids(imgsetid_list)
    # HACK: Get percentage for the annots we currently care about
    aids_list = [(aids) for aids in aids_list_]
    return aids_list


@register_ibs_method
@accessor_decors.getter_1to1
def get_imageset_fraction_imgs_reviewed(ibs, imgsetid_list):
    gids_list = ibs.get_imageset_gids(imgsetid_list)
    flags_list = ibs.unflat_map(ibs.get_image_reviewed, gids_list)
    fraction_imgs_reviewed_list = [
        None if len(flags) == 0 else sum(flags) / len(flags) for flags in flags_list
    ]
    return fraction_imgs_reviewed_list


def _percent_str(pcnt):
    return 'undef' if pcnt is None else '%06.2f %%' % (pcnt * 100,)


@register_ibs_method
@accessor_decors.getter_1to1
@accessor_decors.cache_getter(
    const.IMAGESET_TABLE, 'percent_names_with_exemplar_str', debug=False
)  # HACK
def get_imageset_percent_names_with_exemplar_str(ibs, imgsetid_list):
    fraction_exemplared_names_list = ibs.get_imageset_fraction_names_with_exemplar(
        imgsetid_list
    )
    percent_exemplared_names_list_str = list(
        map(_percent_str, fraction_exemplared_names_list)
    )
    return percent_exemplared_names_list_str


@register_ibs_method
@accessor_decors.getter_1to1
@accessor_decors.cache_getter(
    const.IMAGESET_TABLE, 'percent_imgs_reviewed_str', debug=False
)  # HACK
def get_imageset_percent_imgs_reviewed_str(ibs, imgsetid_list):
    fraction_imgs_reviewed_list = ibs.get_imageset_fraction_imgs_reviewed(imgsetid_list)
    percent_imgs_reviewed_str_list = list(map(_percent_str, fraction_imgs_reviewed_list))
    return percent_imgs_reviewed_str_list


@register_ibs_method
@accessor_decors.getter_1to1
@accessor_decors.cache_getter(
    const.IMAGESET_TABLE, 'percent_annotmatch_reviewed_str', debug=False
)  # HACK
def get_imageset_percent_annotmatch_reviewed_str(ibs, imgsetid_list):
    fraction_annotmatch_reviewed_list = ibs.get_imageset_fraction_annotmatch_reviewed(
        imgsetid_list
    )
    percent_annotmach_reviewed_str_list = list(
        map(_percent_str, fraction_annotmatch_reviewed_list)
    )
    return percent_annotmach_reviewed_str_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/imageset/num/image/', methods=['GET'])
def get_imageset_num_gids(ibs, imgsetid_list):
    r"""
    Returns:
        nGids_list (list): number of images in each imageset

    RESTful:
        Method: GET
        URL:    /api/imageset/num/image/
    """
    nGids_list = list(map(len, ibs.get_imageset_gids(imgsetid_list)))
    return nGids_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/imageset/num/annot/', methods=['GET'])
def get_imageset_num_aids(ibs, imgsetid_list):
    r"""
    Returns:
        nGids_list (list): number of images in each imageset

    RESTful:
        Method: GET
        URL:    /api/imageset/num/annot/
    """
    nAids_list = list(map(len, ibs.get_imageset_aids(imgsetid_list)))
    return nAids_list


@register_ibs_method
@accessor_decors.getter_1toM
@register_api('/api/imageset/annot/rowid/', methods=['GET'])
def get_imageset_aids(ibs, imgsetid_list):
    r"""
    Returns:
        aids_list (list):  a list of list of aids in each imageset

    RESTful:
        Method: GET
        URL:    /api/imageset/annot/rowid/

    Args:
        ibs (IBEISController):  wbia controller object
        imgsetid_list (list):

    Returns:
        list: aids_list

    CommandLine:
        python -m wbia.control.manual_imageset_funcs --test-get_imageset_aids

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.control.manual_imageset_funcs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> ibs.delete_imagesets(ibs.get_valid_imgsetids())
        >>> ibs.compute_occurrences(config={'use_gps': False, 'seconds_thresh': 600})
        >>> imgsetid_list = ibs.get_valid_imgsetids()
        >>> aids_list = get_imageset_aids(ibs, imgsetid_list)
        >>> result = ('aids_list = %s' % (str(aids_list),))
        >>> print(result)
    """
    gids_list = ibs.get_imageset_gids(imgsetid_list)
    aids_list_ = ibs.unflat_map(ibs.get_image_aids, gids_list)
    aids_list = list(map(ut.flatten, aids_list_))
    # logger.info('get_imageset_aids')
    # logger.info('imgsetid_list = %r' % (imgsetid_list,))
    # logger.info('gids_list = %r' % (gids_list,))
    # logger.info('aids_list_ = %r' % (aids_list_,))
    # logger.info('aids_list = %r' % (aids_list,))
    return aids_list


@register_ibs_method
@accessor_decors.getter_1toM
@register_api('/api/imageset/annot/uuid/', methods=['GET'])
def get_imageset_uuids(ibs, imgsetid_list):
    r"""
    Args:
        ibs (IBEISController):  wbia controller object
        imgsetid_list (list):

    Returns:
        list: annot_uuids_list

    RESTful:
        Method: GET
        URL:    /api/imageset/annot/uuid/

    CommandLine:
        python -m wbia.control.manual_imageset_funcs --test-get_imageset_aids

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.control.manual_imageset_funcs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> ibs.delete_imagesets(ibs.get_valid_imgsetids())
        >>> ibs.compute_occurrences(config={'use_gps': False, 'seconds_thresh': 600})
        >>> imgsetid_list = ibs.get_valid_imgsetids()
        >>> aids_list = get_imageset_aids(ibs, imgsetid_list)
        >>> result = ('aids_list = %s' % (str(aids_list),))
        >>> print(result)
    """
    aids_list = ibs.get_imageset_aids(imgsetid_list)
    annot_uuids_list = [ibs.get_annot_uuids(aid_list) for aid_list in aids_list]
    return annot_uuids_list


@register_ibs_method
@accessor_decors.getter_1toM
@accessor_decors.cache_getter(const.IMAGESET_TABLE, 'image_rowids')
@register_api('/api/imageset/image/rowid/', methods=['GET'])
@profile
def get_imageset_gids(ibs, imgsetid_list):
    r"""
    Returns:
        gids_list (list):  a list of list of gids in each imageset

    RESTful:
        Method: GET
        URL:    /api/imageset/image/rowid/
    """
    # FIXME: MAKE SQL-METHOD FOR NON-ROWID GETTERS
    NEW_INDEX_HACK = True
    if NEW_INDEX_HACK:
        # FIXME: This index should when the database is defined.
        # Ensure that an index exists on the image column of the annotation table
        ibs.db.connection.execute(
            """
            CREATE INDEX IF NOT EXISTS gids_to_gs ON {GSG_RELATION_TABLE} (imageset_rowid);
            """.format(
                GSG_RELATION_TABLE=const.GSG_RELATION_TABLE
            )
        ).fetchall()
    gids_list = ibs.db.get(
        const.GSG_RELATION_TABLE,
        ('image_rowid',),
        imgsetid_list,
        id_colname='imageset_rowid',
        unpack_scalars=False,
    )
    # logger.info('get_imageset_gids')
    # logger.info('imgsetid_list = %r' % (imgsetid_list,))
    # logger.info('gids_list = %r' % (gids_list,))
    return gids_list


@register_ibs_method
@accessor_decors.getter_1toM
@register_api('/api/imageset/image/uuid/', methods=['GET'])
def get_imageset_image_uuids(ibs, imgsetid_list):
    r"""
    Returns:
        gids_list (list):  a list of list of gids in each imageset

    RESTful:
        Method: GET
        URL:    /api/imageset/image/uuid/
    """
    gids_list = ibs.get_imageset_gids(imgsetid_list)
    image_uuid_list = [ibs.get_image_uuids(gid_list) for gid_list in gids_list]
    return image_uuid_list


@register_ibs_method
# @register_api('/api/imageset/gsgrids/', methods=['GET'])
def get_imageset_gsgrids(ibs, imgsetid_list=None, gid_list=None):
    r"""
    Returns:
        list_ (list):  a list of imageset-image-relationship rowids for each encouterid
    """
    # WEIRD FUNCTION FIXME
    assert (
        imgsetid_list is not None or gid_list is not None
    ), 'Either imgsetid_list or gid_list must be None'
    if imgsetid_list is not None and gid_list is None:
        # TODO: Group type
        params_iter = ((imgsetid,) for imgsetid in imgsetid_list)
        where_clause = 'imageset_rowid=?'
        # list of relationships for each imageset
        gsgrids_list = ibs.db.get_where(
            const.GSG_RELATION_TABLE,
            ('gsgr_rowid',),
            params_iter,
            where_clause,
            unpack_scalars=False,
        )
    elif gid_list is not None and imgsetid_list is None:
        # TODO: Group type
        params_iter = ((gid,) for gid in gid_list)
        where_clause = 'image_rowid=?'
        # list of relationships for each imageset
        gsgrids_list = ibs.db.get_where(
            const.GSG_RELATION_TABLE,
            ('gsgr_rowid',),
            params_iter,
            where_clause,
            unpack_scalars=False,
        )
    else:
        # TODO: Group type
        params_iter = ((imgsetid, gid) for imgsetid, gid in zip(imgsetid_list, gid_list))
        where_clause = 'imageset_rowid=? AND image_rowid=?'
        # list of relationships for each imageset
        gsgrids_list = ibs.db.get_where(
            const.GSG_RELATION_TABLE,
            ('gsgr_rowid',),
            params_iter,
            where_clause,
            unpack_scalars=False,
        )
    return gsgrids_list


@register_ibs_method
@accessor_decors.getter_1toM
@register_api('/api/imageset/name/rowid/', methods=['GET'])
def get_imageset_nids(ibs, imgsetid_list):
    r"""
    Returns:
        list_ (list):  a list of list of known nids in each imageset

    CommandLine:
        python -m wbia.control.manual_imageset_funcs --test-get_imageset_nids

    RESTful:
        Method: GET
        URL:    /api/imageset/name/rowid/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_imageset_funcs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb('testdb1')
        >>> ibs.delete_imagesets(ibs.get_valid_imgsetids())
        >>> ibs.compute_occurrences(config={'use_gps': False, 'seconds_thresh': 600})
        >>> imgsetid_list = ibs.get_valid_imgsetids()
        >>> nids_list = ibs.get_imageset_nids(imgsetid_list)
        >>> result = nids_list
        >>> print(result)
        [[1, 2, 3], [4, 5, 6, 7]]
    """
    # FIXME: SLOW
    aids_list = ibs.get_imageset_aids(imgsetid_list)
    nids_list = ibs.unflat_map(ibs.get_annot_name_rowids, aids_list)
    # nids_list_ = [[nid[0] for nid in nids if len(nid) > 0] for nids in nids_list]
    # Remove any unknown anmes
    nids_list = [[nid for nid in nids if nid > 0] for nids in nids_list]

    nids_list = list(map(ut.unique_ordered, nids_list))
    # logger.info('get_imageset_nids')
    # logger.info('imgsetid_list = %r' % (imgsetid_list,))
    # logger.info('aids_list = %r' % (aids_list,))
    # logger.info('nids_list_ = %r' % (nids_list_,))
    # logger.info('nids_list = %r' % (nids_list,))
    return nids_list


@register_ibs_method
@accessor_decors.getter_1toM
@register_api('/api/imageset/name/uuid/', methods=['GET'])
def get_imageset_name_uuids(ibs, imgsetid_list):
    r"""
    Returns:
        name_uuid_list (list):  a list of list of known name uuids in each imageset

    CommandLine:
        python -m wbia.control.manual_imageset_funcs --test-get_imageset_name_uuids

    RESTful:
        Method: GET
        URL:    /api/imageset/name/uuid/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_imageset_funcs import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb('testdb1')
        >>> ibs.delete_imagesets(ibs.get_valid_imgsetids())
        >>> ibs.compute_occurrences(config={'use_gps': False, 'seconds_thresh': 600})
        >>> imgsetid_list = ibs.get_valid_imgsetids()
        >>> nids_list = ibs.get_imageset_nids(imgsetid_list)
        >>> result = nids_list
        >>> print(result)
        [[1, 2, 3], [4, 5, 6, 7]]
    """
    nids_list = ibs.get_imageset_nids(imgsetid_list)
    name_uuid_list = [ibs.get_name_uuids(nid_list) for nid_list in nids_list]
    return name_uuid_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/imageset/uuid/', methods=['GET'])
def get_imageset_uuid(ibs, imgsetid_list):
    r"""
    Returns:
        list_ (list): imageset_uuid of each imgsetid in imgsetid_list

    RESTful:
        Method: GET
        URL:    /api/imageset/uuid/
    """
    # FIXME: MAKE SQL-METHOD FOR NON-ROWID GETTERS
    encuuid_list = ibs.db.get(
        const.IMAGESET_TABLE,
        ('imageset_uuid',),
        imgsetid_list,
        id_colname='imageset_rowid',
    )
    return encuuid_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/imageset/text/', methods=['GET'])
def get_imageset_text(ibs, imgsetid_list):
    r"""
    Returns:
        list_ (list): imageset_text of each imgsetid in imgsetid_list

    RESTful:
        Method: GET
        URL:    /api/imageset/text/
    """
    # FIXME: MAKE SQL-METHOD FOR NON-ROWID GETTERS
    imagesettext_list = ibs.db.get(
        const.IMAGESET_TABLE,
        ('imageset_text',),
        imgsetid_list,
        id_colname='imageset_rowid',
    )
    return imagesettext_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/imageset/rowid/uuid/', methods=['GET'])
def get_imageset_imgsetids_from_uuid(ibs, uuid_list):
    r"""
    Returns:
        list_ (list): a list of imgsetids corresponding to each imageset imagesettext
    #FIXME: make new naming scheme for non-primary-key-getters
    get_imageset_imgsetids_from_text_from_text

    RESTful:
        Method: GET
        URL:    /api/imageset/rowid/uuid/
    """
    imgsetid_list = ibs.db.get(
        const.IMAGESET_TABLE, ('imageset_rowid',), uuid_list, id_colname='imageset_uuid'
    )
    return imgsetid_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/imageset/rowid/text/', methods=['GET'])
def get_imageset_imgsetids_from_text(ibs, imagesettext_list, ensure=True):
    r"""
    Returns:
        list_ (list): a list of imgsetids corresponding to each imageset imagesettext
    #FIXME: make new naming scheme for non-primary-key-getters
    get_imageset_imgsetids_from_text_from_text

    RESTful:
        Method: GET
        URL:    /api/imageset/rowid/text/
    """
    if ensure:
        imgsetid_list = ibs.add_imagesets(imagesettext_list)
    else:
        # FIXME: MAKE SQL-METHOD FOR NON-ROWID GETTERS
        imgsetid_list = ibs.db.get(
            const.IMAGESET_TABLE,
            ('imageset_rowid',),
            imagesettext_list,
            id_colname='imageset_text',
        )
    return imgsetid_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/imageset/note/', methods=['GET'])
def get_imageset_note(ibs, imgsetid_list):
    r"""
    Returns:
        list_ (list): imageset_note of each imgsetid in imgsetid_list

    RESTful:
        Method: GET
        URL:    /api/imageset/note/
    """
    # FIXME: MAKE SQL-METHOD FOR NON-ROWID GETTERS
    encnote_list = ibs.db.get(
        const.IMAGESET_TABLE,
        ('imageset_note',),
        imgsetid_list,
        id_colname='imageset_rowid',
    )
    return encnote_list


@register_ibs_method
@accessor_decors.deleter
@register_api('/api/imageset/', methods=['DELETE'])
def delete_imagesets(ibs, imgsetid_list):
    r"""
    Removes imagesets and thier relationships (images are not effected)

    RESTful:
        Method: DELETE
        URL:    /api/imageset/
    """
    # Optimization hack, less SQL calls
    # gsgrid_list = ut.flatten(ibs.get_imageset_gsgrids(imgsetid_list=imgsetid_list))
    # ibs.db.delete_rowids(const.GSG_RELATION_TABLE, gsgrid_list)
    # ibs.db.delete(const.GSG_RELATION_TABLE, imgsetid_list, id_colname='imageset_rowid')
    if ut.VERBOSE:
        logger.info('[ibs] deleting %d imagesets' % len(imgsetid_list))
    ibs.delete_gsgr_imageset_relations(imgsetid_list)
    ibs.db.delete_rowids(const.IMAGESET_TABLE, imgsetid_list)


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/imageset/time/posix/end/', methods=['GET'])
def get_imageset_end_time_posix(ibs, imageset_rowid_list):
    r"""
    imageset_end_time_posix_list <- imageset.imageset_end_time_posix[imageset_rowid_list]

    gets data from the "native" column "imageset_end_time_posix" in the "imageset" table

    Args:
        imageset_rowid_list (list):

    Returns:
        list: imageset_end_time_posix_list

    TemplateInfo:
        Tgetter_table_column
        col = imageset_end_time_posix
        tbl = imageset

    RESTful:
        Method: GET
        URL:    /api/imageset/time/posix/end/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_imageset_funcs import *  # NOQA
        >>> ibs, config2_ = testdata_ibs()
        >>> imageset_rowid_list = ibs._get_all_imageset_rowids()
        >>> imageset_end_time_posix_list = ibs.get_imageset_end_time_posix(imageset_rowid_list)
        >>> assert len(imageset_rowid_list) == len(imageset_end_time_posix_list)
    """
    id_iter = imageset_rowid_list
    colnames = (IMAGESET_END_TIME_POSIX,)
    imageset_end_time_posix_list = ibs.db.get(
        const.IMAGESET_TABLE, colnames, id_iter, id_colname='rowid'
    )
    return imageset_end_time_posix_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/imageset/gps/lat/', methods=['GET'], __api_plural_check__=False)
def get_imageset_gps_lats(ibs, imageset_rowid_list):
    r"""
    imageset_gps_lat_list <- imageset.imageset_gps_lat[imageset_rowid_list]

    gets data from the "native" column "imageset_gps_lat" in the "imageset" table

    Args:
        imageset_rowid_list (list):

    Returns:
        list: imageset_gps_lat_list

    TemplateInfo:
        Tgetter_table_column
        col = imageset_gps_lat
        tbl = imageset

    RESTful:
        Method: GET
        URL:    /api/imageset/gps/lat/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_imageset_funcs import *  # NOQA
        >>> ibs, config2_ = testdata_ibs()
        >>> imageset_rowid_list = ibs._get_all_imageset_rowids()
        >>> imageset_gps_lat_list = ibs.get_imageset_gps_lats(imageset_rowid_list)
        >>> assert len(imageset_rowid_list) == len(imageset_gps_lat_list)
    """
    id_iter = imageset_rowid_list
    colnames = (IMAGESET_GPS_LAT,)
    imageset_gps_lat_list = ibs.db.get(
        const.IMAGESET_TABLE, colnames, id_iter, id_colname='rowid'
    )
    return imageset_gps_lat_list


@register_ibs_method
@register_api('/api/imageset/info/', methods=['PUT'])
def update_imageset_info(ibs, imageset_rowid_list, **kwargs):
    r"""
    sets start and end time for imagesets

    FIXME: should not need to bulk update, should be handled as it goes

    RESTful:
        Method: PUT
        URL:    /api/imageset/info/

    Example:
        >>> # DISABLE_DOCTEST
        >>> imageset_rowid_list = ibs.get_valid_imgsetids()
    """
    gids_list_ = ibs.get_imageset_gids(imageset_rowid_list)
    hasgids_list = [len(gids) > 0 for gids in gids_list_]
    gids_list = ut.compress(gids_list_, hasgids_list)
    imgsetid_list = ut.compress(imageset_rowid_list, hasgids_list)
    unixtimes_list = [
        ibs.get_image_unixtime(gid_list, **kwargs) for gid_list in gids_list
    ]
    # TODO: replace -1's with nans and do nanmin
    imageset_start_time_posix_list = [min(unixtimes) for unixtimes in unixtimes_list]
    imageset_end_time_posix_list = [max(unixtimes) for unixtimes in unixtimes_list]
    ibs.set_imageset_start_time_posix(imgsetid_list, imageset_start_time_posix_list)
    ibs.set_imageset_end_time_posix(imgsetid_list, imageset_end_time_posix_list)


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/imageset/gps/lon/', methods=['GET'], __api_plural_check__=False)
def get_imageset_gps_lons(ibs, imageset_rowid_list):
    r"""
    imageset_gps_lon_list <- imageset.imageset_gps_lon[imageset_rowid_list]

    gets data from the "native" column "imageset_gps_lon" in the "imageset" table

    Args:
        imageset_rowid_list (list):

    Returns:
        list: imageset_gps_lon_list

    TemplateInfo:
        Tgetter_table_column
        col = imageset_gps_lon
        tbl = imageset

    RESTful:
        Method: GET
        URL:    /api/imageset/gps/lon/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_imageset_funcs import *  # NOQA
        >>> ibs, config2_ = testdata_ibs()
        >>> imageset_rowid_list = ibs._get_all_imageset_rowids()
        >>> imageset_gps_lon_list = ibs.get_imageset_gps_lons(imageset_rowid_list)
        >>> assert len(imageset_rowid_list) == len(imageset_gps_lon_list)
    """
    id_iter = imageset_rowid_list
    colnames = (IMAGESET_GPS_LON,)
    imageset_gps_lon_list = ibs.db.get(
        const.IMAGESET_TABLE, colnames, id_iter, id_colname='rowid'
    )
    return imageset_gps_lon_list


@register_ibs_method
@accessor_decors.getter_1to1
# @register_api('/api/imageset/note/', methods=['GET'])
def get_imageset_notes(ibs, imageset_rowid_list):
    r"""
    imageset_note_list <- imageset.imageset_note[imageset_rowid_list]

    gets data from the "native" column "imageset_note" in the "imageset" table

    Args:
        imageset_rowid_list (list):

    Returns:
        list: imageset_note_list

    TemplateInfo:
        Tgetter_table_column
        col = imageset_note
        tbl = imageset

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_imageset_funcs import *  # NOQA
        >>> ibs, config2_ = testdata_ibs()
        >>> imageset_rowid_list = ibs._get_all_imageset_rowids()
        >>> imageset_note_list = ibs.get_imageset_notes(imageset_rowid_list)
        >>> assert len(imageset_rowid_list) == len(imageset_note_list)
    """
    id_iter = imageset_rowid_list
    colnames = (IMAGESET_NOTE,)
    imageset_note_list = ibs.db.get(
        const.IMAGESET_TABLE, colnames, id_iter, id_colname='rowid'
    )
    return imageset_note_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/imageset/metadata/', methods=['GET'])
def get_imageset_metadata(ibs, imageset_rowid_list, return_raw=False):
    r"""
    Returns:
        list_ (list): imageset metadata dictionary

    RESTful:
        Method: GET
        URL:    /api/imageset/metadata/
    """
    metadata_str_list = ibs.db.get(
        const.IMAGESET_TABLE, ('imageset_metadata_json',), imageset_rowid_list
    )
    metadata_list = []
    for metadata_str in metadata_str_list:
        if metadata_str in [None, '']:
            metadata_dict = {}
        else:
            metadata_dict = metadata_str if return_raw else ut.from_json(metadata_str)
        metadata_list.append(metadata_dict)
    return metadata_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/imageset/occurrence/', methods=['GET'])
def get_imageset_occurrence_flags(ibs, imageset_rowid_list):
    r"""
    imageset_occurrence_flag_list <- imageset.imageset_occurrence_flag[imageset_rowid_list]

    gets data from the "native" column "imageset_occurrence_flag" in the "imageset" table

    Args:
        imageset_rowid_list (list):

    Returns:
        list: imageset_occurrence_flag_list

    TemplateInfo:
        Tgetter_table_column
        col = imageset_occurrence_flag
        tbl = imageset

    RESTful:
        Method: GET
        URL:    /api/imageset/occurrence/

    CommandLine:
        python -m wbia.control.manual_imageset_funcs --test-get_imageset_occurrence_flags

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_imageset_funcs import *  # NOQA
        >>> ibs, config2_ = testdata_ibs()
        >>> imageset_rowid_list = ibs._get_all_imageset_rowids()
        >>> imageset_occurrence_flag_list = ibs.get_imageset_occurrence_flags(imageset_rowid_list)
        >>> assert len(imageset_rowid_list) == len(imageset_occurrence_flag_list)
    """
    id_iter = imageset_rowid_list
    colnames = (IMAGESET_OCCURRENCE_FLAG,)
    imageset_occurrence_flag_list = ibs.db.get(
        const.IMAGESET_TABLE, colnames, id_iter, id_colname='rowid'
    )
    return imageset_occurrence_flag_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/imageset/processed/', methods=['GET'])
def get_imageset_processed_flags(ibs, imageset_rowid_list):
    r"""
    imageset_processed_flag_list <- imageset.imageset_processed_flag[imageset_rowid_list]

    gets data from the "native" column "imageset_processed_flag" in the "imageset" table

    Args:
        imageset_rowid_list (list):

    Returns:
        list: imageset_processed_flag_list

    TemplateInfo:
        Tgetter_table_column
        col = imageset_processed_flag
        tbl = imageset

    RESTful:
        Method: GET
        URL:    /api/imageset/processed/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_imageset_funcs import *  # NOQA
        >>> ibs, config2_ = testdata_ibs()
        >>> imageset_rowid_list = ibs._get_all_imageset_rowids()
        >>> imageset_processed_flag_list = ibs.get_imageset_processed_flags(imageset_rowid_list)
        >>> assert len(imageset_rowid_list) == len(imageset_processed_flag_list)
    """
    id_iter = imageset_rowid_list
    colnames = (IMAGESET_PROCESSED_FLAG,)
    imageset_processed_flag_list = ibs.db.get(
        const.IMAGESET_TABLE, colnames, id_iter, id_colname='rowid'
    )
    return imageset_processed_flag_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/imageset/shipped/', methods=['GET'])
def get_imageset_shipped_flags(ibs, imageset_rowid_list):
    r"""
    imageset_shipped_flag_list <- imageset.imageset_shipped_flag[imageset_rowid_list]

    gets data from the "native" column "imageset_shipped_flag" in the "imageset" table

    Args:
        imageset_rowid_list (list):

    Returns:
        list: imageset_shipped_flag_list

    TemplateInfo:
        Tgetter_table_column
        col = imageset_shipped_flag
        tbl = imageset

    RESTful:
        Method: GET
        URL:    /api/imageset/shipped/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_imageset_funcs import *  # NOQA
        >>> ibs, config2_ = testdata_ibs()
        >>> imageset_rowid_list = ibs._get_all_imageset_rowids()
        >>> imageset_shipped_flag_list = ibs.get_imageset_shipped_flags(imageset_rowid_list)
        >>> assert len(imageset_rowid_list) == len(imageset_shipped_flag_list)
    """
    id_iter = imageset_rowid_list
    colnames = (IMAGESET_SHIPPED_FLAG,)
    imageset_shipped_flag_list = ibs.db.get(
        const.IMAGESET_TABLE, colnames, id_iter, id_colname='rowid'
    )
    return imageset_shipped_flag_list


@register_ibs_method
@accessor_decors.getter_1to1
@register_api('/api/imageset/time/posix/start/', methods=['GET'])
def get_imageset_start_time_posix(ibs, imageset_rowid_list):
    r"""
    imageset_start_time_posix_list <- imageset.imageset_start_time_posix[imageset_rowid_list]

    gets data from the "native" column "imageset_start_time_posix" in the "imageset" table

    Args:
        imageset_rowid_list (list):

    Returns:
        list: imageset_start_time_posix_list

    TemplateInfo:
        Tgetter_table_column
        col = imageset_start_time_posix
        tbl = imageset

    RESTful:
        Method: GET
        URL:    /api/imageset/time/posix/start/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_imageset_funcs import *  # NOQA
        >>> ibs, config2_ = testdata_ibs()
        >>> imageset_rowid_list = ibs._get_all_imageset_rowids()
        >>> imageset_start_time_posix_list = ibs.get_imageset_start_time_posix(imageset_rowid_list)
        >>> assert len(imageset_rowid_list) == len(imageset_start_time_posix_list)
    """
    id_iter = imageset_rowid_list
    colnames = (IMAGESET_START_TIME_POSIX,)
    imageset_start_time_posix_list = ibs.db.get(
        const.IMAGESET_TABLE, colnames, id_iter, id_colname='rowid'
    )
    return imageset_start_time_posix_list


@register_ibs_method
@accessor_decors.getter
@register_api('/api/imageset/duration/', methods=['GET'])
def get_imageset_duration(ibs, imageset_rowid_list):
    r"""
    gets the imageset's duration

    Args:
        imageset_rowid_list (list):

    Returns:
        list: imageset_duration

    RESTful:
        Method: GET
        URL:    /api/imageset/duration/
    """

    def _process(start, end):
        if start is None or end is None:
            return 'None'
        seconds_in_day = 60 * 60 * 24
        days = 0
        duration = int(end - start)
        if duration >= seconds_in_day:
            days = duration // seconds_in_day
            duration = duration % seconds_in_day
        duration_str = time.strftime('%H:%M:%S', time.gmtime(duration))
        if days > 0:
            duration_str = '%d days, %s' % (days, duration_str)
        return duration_str

    import time

    start_time_list = ibs.get_imageset_start_time_posix(imageset_rowid_list)
    end_time_list = ibs.get_imageset_end_time_posix(imageset_rowid_list)
    zipped = zip(start_time_list, end_time_list)
    duration_list = [_process(start, end) for start, end in zipped]
    return duration_list


@register_ibs_method
@accessor_decors.setter
@register_api('/api/imageset/time/posix/end/', methods=['PUT'])
def set_imageset_end_time_posix(ibs, imageset_rowid_list, imageset_end_time_posix_list):
    r"""
    imageset_end_time_posix_list -> imageset.imageset_end_time_posix[imageset_rowid_list]

    Args:
        imageset_rowid_list
        imageset_end_time_posix_list

    TemplateInfo:
        Tsetter_native_column
        tbl = imageset
        col = imageset_end_time_posix

    RESTful:
        Method: PUT
        URL:    /api/imageset/time/posix/end/
    """
    id_iter = imageset_rowid_list
    colnames = (IMAGESET_END_TIME_POSIX,)
    ibs.db.set(const.IMAGESET_TABLE, colnames, imageset_end_time_posix_list, id_iter)


@register_ibs_method
@accessor_decors.setter
@register_api('/api/imageset/gps/lat/', methods=['PUT'], __api_plural_check__=False)
def set_imageset_gps_lats(ibs, imageset_rowid_list, imageset_gps_lat_list):
    r"""
    imageset_gps_lat_list -> imageset.imageset_gps_lat[imageset_rowid_list]

    Args:
        imageset_rowid_list
        imageset_gps_lat_list

    TemplateInfo:
        Tsetter_native_column
        tbl = imageset
        col = imageset_gps_lat

    RESTful:
        Method: PUT
        URL:    /api/imageset/gps/lat/
    """
    id_iter = imageset_rowid_list
    colnames = (IMAGESET_GPS_LAT,)
    ibs.db.set(const.IMAGESET_TABLE, colnames, imageset_gps_lat_list, id_iter)


@register_ibs_method
@accessor_decors.setter
@register_api('/api/imageset/gps/lon/', methods=['PUT'], __api_plural_check__=False)
def set_imageset_gps_lons(ibs, imageset_rowid_list, imageset_gps_lon_list):
    r"""
    imageset_gps_lon_list -> imageset.imageset_gps_lon[imageset_rowid_list]

    Args:
        imageset_rowid_list
        imageset_gps_lon_list

    TemplateInfo:
        Tsetter_native_column
        tbl = imageset
        col = imageset_gps_lon

    RESTful:
        Method: PUT
        URL:    /api/imageset/gps/lon/
    """
    id_iter = imageset_rowid_list
    colnames = (IMAGESET_GPS_LON,)
    ibs.db.set(const.IMAGESET_TABLE, colnames, imageset_gps_lon_list, id_iter)


@register_ibs_method
@accessor_decors.setter
@register_api('/api/imageset/note/', methods=['PUT'])
def set_imageset_notes(ibs, imageset_rowid_list, imageset_note_list):
    r"""
    imageset_note_list -> imageset.imageset_note[imageset_rowid_list]

    Args:
        imageset_rowid_list
        imageset_note_list

    TemplateInfo:
        Tsetter_native_column
        tbl = imageset
        col = imageset_note
    """
    id_iter = imageset_rowid_list
    colnames = (IMAGESET_NOTE,)
    ibs.db.set(const.IMAGESET_TABLE, colnames, imageset_note_list, id_iter)


@register_ibs_method
@accessor_decors.setter
@register_api('/api/imageset/metadata/', methods=['PUT'])
def set_imageset_metadata(ibs, imageset_rowid_list, metadata_dict_list):
    r"""
    Sets the imageset's metadata using a metadata dictionary

    RESTful:
        Method: PUT
        URL:    /api/imageset/metadata/

    """
    id_iter = ((gid,) for gid in imageset_rowid_list)
    metadata_str_list = []
    for metadata_dict in metadata_dict_list:
        metadata_str = ut.to_json(metadata_dict)
        metadata_str_list.append(metadata_str)
    val_list = ((metadata_str,) for metadata_str in metadata_str_list)
    ibs.db.set(const.IMAGESET_TABLE, ('imageset_metadata_json',), val_list, id_iter)


@register_ibs_method
@accessor_decors.setter
@register_api('/api/imageset/occurrence/', methods=['PUT'])
def set_imageset_occurrence_flags(
    ibs, imageset_rowid_list, imageset_occurrence_flag_list
):
    r"""
    imageset_occurrence_flag_list -> imageset.imageset_occurrence_flag[imageset_rowid_list]

    Args:
        imageset_rowid_list
        imageset_occurrence_flag_list

    TemplateInfo:
        Tsetter_native_column
        tbl = imageset
        col = imageset_occurrence_flag

    RESTful:
        Method: PUT
        URL:    /api/imageset/occurrence/
    """
    id_iter = imageset_rowid_list
    colnames = (IMAGESET_OCCURRENCE_FLAG,)
    val_iter = ((occurrence_flag,) for occurrence_flag in imageset_occurrence_flag_list)
    ibs.db.set(const.IMAGESET_TABLE, colnames, val_iter, id_iter)


@register_ibs_method
@accessor_decors.setter
@register_api('/api/imageset/processed/', methods=['PUT'])
def set_imageset_processed_flags(ibs, imageset_rowid_list, imageset_processed_flag_list):
    r"""
    imageset_processed_flag_list -> imageset.imageset_processed_flag[imageset_rowid_list]

    Args:
        imageset_rowid_list
        imageset_processed_flag_list

    TemplateInfo:
        Tsetter_native_column
        tbl = imageset
        col = imageset_processed_flag

    RESTful:
        Method: PUT
        URL:    /api/imageset/processed/
    """
    id_iter = imageset_rowid_list
    colnames = (IMAGESET_PROCESSED_FLAG,)
    val_iter = ((processed_flag,) for processed_flag in imageset_processed_flag_list)
    ibs.db.set(const.IMAGESET_TABLE, colnames, val_iter, id_iter)


@register_ibs_method
@accessor_decors.setter
@register_api('/api/imageset/shipped/', methods=['PUT'])
def set_imageset_shipped_flags(ibs, imageset_rowid_list, imageset_shipped_flag_list):
    r"""
    imageset_shipped_flag_list -> imageset.imageset_shipped_flag[imageset_rowid_list]

    Args:
        imageset_rowid_list
        imageset_shipped_flag_list

    TemplateInfo:
        Tsetter_native_column
        tbl = imageset
        col = imageset_shipped_flag

    RESTful:
        Method: PUT
        URL:    /api/imageset/shipped/
    """
    id_iter = imageset_rowid_list
    colnames = (IMAGESET_SHIPPED_FLAG,)
    val_iter = ((shipped_flag,) for shipped_flag in imageset_shipped_flag_list)
    ibs.db.set(const.IMAGESET_TABLE, colnames, val_iter, id_iter)


@register_ibs_method
@accessor_decors.setter
@register_api('/api/imageset/time/posix/start/', methods=['PUT'])
def set_imageset_start_time_posix(
    ibs, imageset_rowid_list, imageset_start_time_posix_list
):
    r"""
    imageset_start_time_posix_list -> imageset.imageset_start_time_posix[imageset_rowid_list]

    Args:
        imageset_rowid_list
        imageset_start_time_posix_list

    TemplateInfo:
        Tsetter_native_column
        tbl = imageset
        col = imageset_start_time_posix

    RESTful:
        Method: PUT
        URL:    /api/imageset/time/posix/start/
    """
    id_iter = imageset_rowid_list
    colnames = (IMAGESET_START_TIME_POSIX,)
    ibs.db.set(const.IMAGESET_TABLE, colnames, imageset_start_time_posix_list, id_iter)


@register_ibs_method
# @accessor_decors.cache_getter(const.IMAGESET_TABLE, IMAGESET_SMART_WAYPOINT_ID)
@register_api('/api/imageset/smart/waypoint/', methods=['GET'])
def get_imageset_smart_waypoint_ids(ibs, imageset_rowid_list):
    r"""
    imageset_smart_waypoint_id_list <- imageset.imageset_smart_waypoint_id[imageset_rowid_list]

    gets data from the "native" column "imageset_smart_waypoint_id" in the "imageset" table

    Args:
        imageset_rowid_list (list):

    Returns:
        list: imageset_smart_waypoint_id_list

    TemplateInfo:
        Tgetter_table_column
        col = imageset_smart_waypoint_id
        tbl = imageset

    RESTful:
        Method: GET
        URL:    /api/imageset/smart/waypoint/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_imageset_funcs import *  # NOQA
        >>> ibs, config2_ = testdata_ibs()
        >>> imageset_rowid_list = ibs._get_all_imageset_rowids()
        >>> imageset_smart_waypoint_id_list = ibs.get_imageset_smart_waypoint_ids(imageset_rowid_list)
        >>> assert len(imageset_rowid_list) == len(imageset_smart_waypoint_id_list)
    """
    id_iter = imageset_rowid_list
    colnames = (IMAGESET_SMART_WAYPOINT_ID,)
    imageset_smart_waypoint_id_list = ibs.db.get(
        const.IMAGESET_TABLE, colnames, id_iter, id_colname='rowid'
    )
    return imageset_smart_waypoint_id_list


@register_ibs_method
# @accessor_decors.cache_getter(const.IMAGESET_TABLE, IMAGESET_SMART_XML_FNAME)
@register_api('/api/imageset/smart/xml/file/name/', methods=['GET'])
def get_imageset_smart_xml_fnames(ibs, imageset_rowid_list):
    r"""
    imageset_smart_xml_fname_list <- imageset.imageset_smart_xml_fname[imageset_rowid_list]

    gets data from the "native" column "imageset_smart_xml_fname" in the "imageset" table

    Args:
        imageset_rowid_list (list):

    Returns:
        list: imageset_smart_xml_fname_list

    TemplateInfo:
        Tgetter_table_column
        col = imageset_smart_xml_fname
        tbl = imageset

    RESTful:
        Method: GET
        URL:    /api/imageset/smart/xml/file/name/

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.control.manual_imageset_funcs import *  # NOQA
        >>> ibs, config2_ = testdata_ibs()
        >>> imageset_rowid_list = ibs._get_all_imageset_rowids()
        >>> imageset_smart_xml_fname_list = ibs.get_imageset_smart_xml_fnames(imageset_rowid_list)
        >>> assert len(imageset_rowid_list) == len(imageset_smart_xml_fname_list)
    """
    id_iter = imageset_rowid_list
    colnames = (IMAGESET_SMART_XML_FNAME,)
    imageset_smart_xml_fname_list = ibs.db.get(
        const.IMAGESET_TABLE, colnames, id_iter, id_colname='rowid'
    )
    return imageset_smart_xml_fname_list


@register_ibs_method
# @accessor_decors.cache_getter(const.IMAGESET_TABLE, IMAGESET_SMART_XML_FNAME)
@register_api('/api/imageset/smart/xml/file/content/', methods=['GET'])
def get_imageset_smart_xml_contents(ibs, imageset_rowid_list):
    from os.path import join, exists

    imageset_smart_xml_fname_list = ibs.get_imageset_smart_xml_fnames(imageset_rowid_list)
    content_list = []
    smart_patrol_dir = ibs.get_smart_patrol_dir()
    for imageset_smart_xml_fname in imageset_smart_xml_fname_list:
        if imageset_smart_xml_fname is None:
            content_list.append(None)
        else:
            imageset_smart_xml_fpath = join(smart_patrol_dir, imageset_smart_xml_fname)
            if exists(imageset_smart_xml_fpath):
                with open(imageset_smart_xml_fpath, 'r') as imageset_smart_xml:
                    content = imageset_smart_xml.read()
                    content_list.append(content)
            else:
                content_list.append(None)
    return content_list


@register_ibs_method
@register_api('/api/imageset/smart/waypoint/', methods=['PUT'])
def set_imageset_smart_waypoint_ids(
    ibs, imageset_rowid_list, imageset_smart_waypoint_id_list
):
    r"""
    imageset_smart_waypoint_id_list -> imageset.imageset_smart_waypoint_id[imageset_rowid_list]

    Args:
        imageset_rowid_list
        imageset_smart_waypoint_id_list

    TemplateInfo:
        Tsetter_native_column
        tbl = imageset
        col = imageset_smart_waypoint_id

    RESTful:
        Method: PUT
        URL:    /api/imageset/smart/waypoint/
    """
    id_iter = imageset_rowid_list
    colnames = (IMAGESET_SMART_WAYPOINT_ID,)
    ibs.db.set(const.IMAGESET_TABLE, colnames, imageset_smart_waypoint_id_list, id_iter)


@register_ibs_method
@register_api('/api/imageset/smart/xml/file/name/', methods=['PUT'])
def set_imageset_smart_xml_fnames(
    ibs, imageset_rowid_list, imageset_smart_xml_fname_list
):
    r"""
    imageset_smart_xml_fname_list -> imageset.imageset_smart_xml_fname[imageset_rowid_list]

    Args:
        imageset_rowid_list
        imageset_smart_xml_fname_list

    TemplateInfo:
        Tsetter_native_column
        tbl = imageset
        col = imageset_smart_xml_fname

    RESTful:
        Method: PUT
        URL:    /api/imageset/smart/xml/fname/
    """
    id_iter = imageset_rowid_list
    colnames = (IMAGESET_SMART_XML_FNAME,)
    ibs.db.set(const.IMAGESET_TABLE, colnames, imageset_smart_xml_fname_list, id_iter)


def testdata_ibs():
    r""""""
    import wbia

    ibs = wbia.opendb('testdb1')
    config2_ = None
    return ibs, config2_



# -*- coding: utf-8 -*-
# developer convenience functions for ibs
import logging
import utool as ut
from six.moves import zip
from wbia import constants as const

(print, rrr, profile) = ut.inject2(__name__, '[duct_tape]')
logger = logging.getLogger('wbia')


def fix_compname_configs(ibs):
    """ duct tape to keep version in check """
    # ibs.MANUAL_CONFIG_SUFFIX = '_MANUAL_'  #+ ut.get_computer_name()
    # ibs.MANUAL_CONFIGID = ibs.add_config(ibs.MANUAL_CONFIG_SUFFIX)
    # We need to fix the manual config suffix to not use computer names anymore

    configid_list = ibs.get_valid_configids()
    cfgsuffix_list = ibs.get_config_suffixes(configid_list)

    ibs.MANUAL_CONFIG_SUFFIX = 'MANUAL_CONFIG'
    ibs.MANUAL_CONFIGID = ibs.add_config(ibs.MANUAL_CONFIG_SUFFIX)

    for rowid, suffix in filter(
        lambda tup: tup[1].startswith('_MANUAL_'), zip(configid_list, cfgsuffix_list)
    ):
        logger.info('EVALUATING: %r, %r' % (rowid, suffix))
        # Fix the tables with bad config_rowids
        ibs.db.executeone(
            """
            UPDATE {AL_RELATION_TABLE}
            SET config_rowid=?
            WHERE config_rowid=?
            """.format(
                **const.__dict__
            ),
            params=(ibs.MANUAL_CONFIGID, rowid),
        )

        # Delete the bad config_suffixes
        ibs.db.executeone(
            """
            DELETE
            FROM {CONFIG_TABLE}
            WHERE config_rowid=?
            """.format(
                **const.__dict__
            ),
            params=(rowid,),
        )


def remove_database_slag(
    ibs,
    delete_empty_names=False,
    delete_empty_imagesets=False,
    delete_annotations_for_missing_images=False,
    delete_image_labels_for_missing_types=False,
    delete_annot_labels_for_missing_types=False,
    delete_chips_for_missing_annotations=False,
    delete_features_for_missing_annotations=False,
    delete_invalid_eg_relations=False,
    delete_invalid_gl_relations=False,
    delete_invalid_al_relations=True,
):
    # ZERO ORDER
    if delete_empty_names:
        ibs.delete_empty_nids()

    if delete_empty_imagesets:
        ibs.delete_empty_imgsetids()

    # FIRST ORDER
    if delete_annotations_for_missing_images:
        ibs.db.executeone(
            """
            DELETE
            FROM {ANNOTATION_TABLE}
            WHERE
                image_rowid NOT IN (SELECT rowid FROM {IMAGE_TABLE})
            """.format(
                **const.__dict__
            )
        )

    if delete_image_labels_for_missing_types:
        ibs.db.executeone(
            """
            DELETE
            FROM {LBLIMAGE_TABLE}
            WHERE
                lbltype_rowid NOT IN (SELECT rowid FROM {LBLTYPE_TABLE})
            """.format(
                **const.__dict__
            )
        )

    if delete_annot_labels_for_missing_types:
        ibs.db.executeone(
            """
            DELETE
            FROM {LBLANNOT_TABLE}
            WHERE
                lbltype_rowid NOT IN (SELECT rowid FROM {LBLTYPE_TABLE})
            """.format(
                **const.__dict__
            )
        )

    # SECOND ORDER
    if delete_chips_for_missing_annotations:
        ibs.db.executeone(
            """
            DELETE
            FROM {CHIP_TABLE}
            WHERE
                annot_rowid NOT IN (SELECT rowid FROM {ANNOTATION_TABLE})
            """.format(
                **const.__dict__
            )
        )
        # OR config_rowid NOT IN (SELECT rowid FROM {CONFIG_TABLE})

    if delete_features_for_missing_annotations:
        ibs.db.executeone(
            """
            DELETE
            FROM {FEATURE_TABLE}
            WHERE
                chip_rowid NOT IN (SELECT rowid FROM {CHIP_TABLE})
            """.format(
                **const.__dict__
            )
        )
        # OR config_rowid NOT IN (SELECT rowid FROM {CONFIG_TABLE})

    if delete_invalid_eg_relations:
        ibs.db.executeone(
            """
            DELETE
            FROM {GSG_RELATION_TABLE}
            WHERE
                image_rowid NOT IN (SELECT rowid FROM {IMAGE_TABLE}) OR
                imageset_rowid NOT IN (SELECT rowid FROM {IMAGESET_TABLE})
            """.format(
                **const.__dict__
            )
        )

    # THIRD ORDER
    if delete_invalid_gl_relations:
        ibs.db.executeone(
            """
            DELETE
            FROM {GL_RELATION_TABLE}
            WHERE
                image_rowid NOT IN (SELECT rowid FROM {IMAGE_TABLE}) OR
                lblimage_rowid NOT IN (SELECT rowid FROM {LBLIMAGE_TABLE})
            """.format(
                **const.__dict__
            )
        )
        # OR config_rowid NOT IN (SELECT rowid FROM {CONFIG_TABLE})

    if delete_invalid_al_relations:
        ibs.db.executeone(
            """
            DELETE
            FROM {AL_RELATION_TABLE}
            WHERE
                annot_rowid NOT IN (SELECT rowid FROM {ANNOTATION_TABLE}) OR
                lblannot_rowid NOT IN (SELECT rowid FROM {LBLANNOT_TABLE})
            """.format(
                **const.__dict__
            )
        )
        # OR config_rowid NOT IN (SELECT rowid FROM {CONFIG_TABLE})


def enforce_unkonwn_name_is_explicit(ibs):
    nid_list = ibs.get_valid_nids()
    text_list = ibs.get_name_texts(nid_list)
    problem_nids = [
        text for text, nid in zip(text_list, nid_list) if text == const.UNKNOWN
    ]
    unknown_aids = ibs.get_name_aids(problem_nids)
    assert len(ut.flatten(unknown_aids)) == 0
    # TODO Take unknown_aids and remove any name relationships to make unknown
    # implicit


def fix_nulled_yaws(ibs):
    aid_list = ibs.get_valid_aids()
    yaw_list = ibs.get_annot_yaws(aid_list)
    valid_list = [yaw == 0.0 for yaw in yaw_list]
    dirty_aid_list = ut.filter_items(aid_list, valid_list)
    logger.info('[duct_tape] Nulling %d annotation yaws' % len(dirty_aid_list))
    ibs.set_annot_viewpoints(dirty_aid_list, [None] * len(dirty_aid_list))



# -*- coding: utf-8 -*-
import sqlite3
import uuid

import numpy as np
import pytest

# We do not explicitly call code in this module because
# importing the following module is execution of the code.
import wbia.dtool._integrate_sqlite3  # noqa


@pytest.fixture
def db():
    with sqlite3.connect(':memory:', detect_types=sqlite3.PARSE_DECLTYPES) as con:
        yield con


np_number_types = (
    np.int8,
    np.int16,
    np.int32,
    np.int64,
    np.uint8,
    np.uint16,
    np.uint32,
    np.uint64,
)


@pytest.mark.parametrize('num_type', np_number_types)
def test_register_numpy_dtypes_ints(db, num_type):
    # The magic takes place in the register_numpy_dtypes function,
    # which is implicitly called on module import.

    # Create a table that uses the type
    db.execute('create table test(x integer)')

    # Insert a uuid value into the table
    insert_value = num_type(8)
    db.execute('insert into test(x) values (?)', (insert_value,))
    # Query for the value
    cur = db.execute('select x from test')
    selected_value = cur.fetchone()[0]
    assert selected_value == insert_value


@pytest.mark.parametrize('num_type', (np.float32, np.float64))
def test_register_numpy_dtypes_floats(db, num_type):
    # The magic takes place in the register_numpy_dtypes function,
    # which is implicitly called on module import.

    # Create a table that uses the type
    db.execute('create table test(x real)')

    # Insert a uuid value into the table
    insert_value = num_type(8.0000008)
    db.execute('insert into test(x) values (?)', (insert_value,))
    # Query for the value
    cur = db.execute('select x from test')
    selected_value = cur.fetchone()[0]
    assert selected_value == insert_value


@pytest.mark.parametrize('type_name', ('numpy', 'ndarray'))
def test_register_numpy(db, type_name):
    # The magic takes place in the register_numpy function,
    # which is implicitly called on module import.

    # Create a table that uses the type
    db.execute(f'create table test(x {type_name})')

    # Insert a numpy array value into the table
    insert_value = np.array([[1, 2, 3], [4, 5, 6]], np.int32)
    db.execute('insert into test(x) values (?)', (insert_value,))
    # Query for the value
    cur = db.execute('select x from test')
    selected_value = cur.fetchone()[0]
    assert (selected_value == insert_value).all()


def test_register_uuid(db):
    # The magic takes place in the register_uuid function,
    # which is implicitly called on module import.

    # Create a table that uses the type
    db.execute('create table test(x uuid)')

    # Insert a uuid value into the table
    insert_value = uuid.uuid4()
    db.execute('insert into test(x) values (?)', (insert_value,))
    # Query for the value
    cur = db.execute('select x from test')
    selected_value = cur.fetchone()[0]
    assert selected_value == insert_value


def test_register_dict(db):
    # The magic takes place in the register_dict function,
    # which is implicitly called on module import.

    # Create a table that uses the type
    db.execute('create table test(x dict)')

    # Insert a dict value into the table
    insert_value = {
        'a': 1,
        'b': 2.2,
        'c': [[1, 2, 3], [4, 5, 6]],
    }
    db.execute('insert into test(x) values (?)', (insert_value,))
    # Query for the value
    cur = db.execute('select x from test')
    selected_value = cur.fetchone()[0]
    for k, v in selected_value.items():
        assert v == insert_value[k]


def test_register_list(db):
    # The magic takes place in the register_list function,
    # which is implicitly called on module import.

    # Create a table that uses the type
    db.execute('create table test(x list)')

    # Insert a list of list value into the table
    insert_value = [[1, 2, 3], [4, 5, 6]]
    db.execute('insert into test(x) values (?)', (insert_value,))
    # Query for the value
    cur = db.execute('select x from test')
    selected_value = cur.fetchone()[0]
    assert selected_value == insert_value
